################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
revigo.py - apply semantic clustering to GO output
==================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

apply semantic clustering ala revigo (:pmid:`21789182`, http://revigo.irb.hr)
to a list of significantly enriched or depleted GO terms.

This script can be used to reduce the redundancy in large lists of
GO results. The principal input is list of GO-terms and associated
P-Values.

Usage
-----

Example::

   python revigo.py

Type::

   python revigo.py --help

for command line help.

Documentation
-------------

In addition to a list of GO terms and their associated p-values (--filename-pvalues)

Additional input files required are:

* a file containing a gene ontology in obo-xml format (--filename-ontology)
* a file mapping gene identifiers to GO terms (--filename-go)

The script will compute a matrix of semantic similarities between GO terms
based on the ``simrel`` measure described in the manuscript by Schlicker et al. (2006)
(pmid:`16776819`).

The script will then remove redundant terms from this matrix following the produce
described in the revigo manuscript (:pmid:`21789182`).

The script will output a file called `cluster.tsv` which contains the clustering results.

It will also output an image in :term:`svg` format of the clustered terms and their
similarity in semantic space. Nodes are coloured by fold change and size is given
by the P-Value.

Additional output files
+++++++++++++++++++++++

Computing the similarity matrix can take same time. The script outputs some intermediate
files that will be re-used on subsequent calls in order to speed up processing. The intermediate 
files are:

    * ancestors.tsv: tab-separated file with all the ancestors of a term
    * termfrequencies.tsv: tab-separated file with all the term frequencies and genes with this term
    * simrel.matrix: the full simrel matrix. This matrix is symmetric.
    * cluster.matrix: the clustered simrel matrix. Only the cluster representatives have been kept.

Quality control
+++++++++++++++

The output of this program is similar but not identical to revigo. The algorithm is not fully described
and the input data are not identical. It seems that simrel values in this script are lower and thus the
clustering is more conservative compared to revigo.

Code
----

This module depends on unreleased code from http://code.google.com/p/gographer for parsing of the
GO graph. The code has been modified to allow parsing of relationship information and record synonyms
and obsolete identifiers. The code has been included into this script until gographer will be released. 

'''

import itertools
import math
import sys
import optparse
import collections
import os
import re
import string

import CGAT.GO as GO
import CGAT.IOTools as IOTools

import CGAT.Experiment as E
import numpy
import networkx
import networkx.algorithms.traversal.depth_first_search as traversal
from xml.sax.handler import ContentHandler
from xml.sax import make_parser
from networkx import DiGraph
from networkx import topological_sort
import cPickle

import matplotlib
matplotlib.use('Agg') 
import matplotlib.pyplot as plt
import pylab

class Tokenizer:
    '''from gographer utils.py'''
    def __init__(self):
        self.pattern_punctuation = re.compile('(\!|\.|\+|\?|\;|\,|\:|\&|\_|\~|\:|\-|\;|\$|\@|\#|\%|\*|\^|\`|\|)')
        self.pattern_numbers = re.compile('[0-9]')
        self.pattern_plurals = re.compile('\'s$')
        self.pattern_possessives = re.compile('s\'$')
        self.pattern_blanks = re.compile('[\t, ,\n,\r,\f,\v,]+')
        self.pattern_others = re.compile('(\'|\"|\/|\\\|\<|\>|\{|\}|\[|\]|\(|\))')
        
        
        self.pattern_words_and_blanks = re.compile('[\W\t\n\r\f\v| ]|\|')
        #self.pattern.word = re.compile('[a-z]+')[^a-zA-Z]

        self.leftchar = {}
        self.leftchar[' ']=0
        for i in string.ascii_letters:
            self.leftchar[i]=0
        #for i in range(10):
            #self.leftchar[str(i)]=0
        #print len(self.leftchar)

            
    def keepNumChr(self, words):
        newWords = ''
        for c in words:
            if self.leftchar.has_key(c):
                newWords +=c
        return newWords
    
                                
    def tokenize_word(self,words):
        before = words
        words = words.lower()
        words = self.pattern_plurals.sub('', words)
        words = self.pattern_possessives.sub('', words)
        words = self.keepNumChr(words)
        words = self.pattern_blanks.sub(' ',words)
        
        #words = string.strip(words)    
        return words

        
    def test(self,words):
        for c in words:    
            if not self.leftchar.has_key(c):
                self.leftchar[c] = 1

                
    def returnUnexpChar(self):
        reV = []
        for k in self.leftchar.keys():
            if self.leftchar[k] != 0:
                reV +=[k]
                
        return reV

class PorterStemmer:
    '''from gographer utils.py'''
    def __init__(self):
        """The main part of the stemming algorithm starts here.
        b is a buffer holding a word to be stemmed. The letters are in b[k0],
        b[k0+1] ... ending at b[k]. In fact k0 = 0 in this demo program. k is
        readjusted downwards as the stemming progresses. Zero termination is
        not in fact used in the algorithm.

        Note that only lower case sequences are stemmed. Forcing to lower case
        should be done before stem(...) is called.
        """

        self.b = ""  # buffer for word to be stemmed
        self.k = 0
        self.k0 = 0
        self.j = 0   # j is a general offset into the string

    def cons(self, i):
        """cons(i) is TRUE <=> b[i] is a consonant."""
        if self.b[i] == 'a' or self.b[i] == 'e' or self.b[i] == 'i' or self.b[i] == 'o' or self.b[i] == 'u':
            return 0
        if self.b[i] == 'y':
            if i == self.k0:
                return 1
            else:
                return (not self.cons(i - 1))
        return 1

    def m(self):
        """m() measures the number of consonant sequences between k0 and j.
        if c is a consonant sequence and v a vowel sequence, and <..>
        indicates arbitrary presence,

           <c><v>       gives 0
           <c>vc<v>     gives 1
           <c>vcvc<v>   gives 2
           <c>vcvcvc<v> gives 3
           ....
        """
        n = 0
        i = self.k0
        while 1:
            if i > self.j:
                return n
            if not self.cons(i):
                break
            i = i + 1
        i = i + 1
        while 1:
            while 1:
                if i > self.j:
                    return n
                if self.cons(i):
                    break
                i = i + 1
            i = i + 1
            n = n + 1
            while 1:
                if i > self.j:
                    return n
                if not self.cons(i):
                    break
                i = i + 1
            i = i + 1

    def vowelinstem(self):
        """vowelinstem() is TRUE <=> k0,...j contains a vowel"""
        for i in range(self.k0, self.j + 1):
            if not self.cons(i):
                return 1
        return 0

    def doublec(self, j):
        """doublec(j) is TRUE <=> j,(j-1) contain a double consonant."""
        if j < (self.k0 + 1):
            return 0
        if (self.b[j] != self.b[j-1]):
            return 0
        return self.cons(j)

    def cvc(self, i):
        """cvc(i) is TRUE <=> i-2,i-1,i has the form consonant - vowel - consonant
        and also if the second c is not w,x or y. this is used when trying to
        restore an e at the end of a short  e.g.

           cav(e), lov(e), hop(e), crim(e), but
           snow, box, tray.
        """
        if i < (self.k0 + 2) or not self.cons(i) or self.cons(i-1) or not self.cons(i-2):
            return 0
        ch = self.b[i]
        if ch == 'w' or ch == 'x' or ch == 'y':
            return 0
        return 1

    def ends(self, s):
        """ends(s) is TRUE <=> k0,...k ends with the string s."""
        length = len(s)
        if s[length - 1] != self.b[self.k]: # tiny speed-up
            return 0
        if length > (self.k - self.k0 + 1):
            return 0
        if self.b[self.k-length+1:self.k+1] != s:
            return 0
        self.j = self.k - length
        return 1

    def setto(self, s):
        """setto(s) sets (j+1),...k to the characters in the string s, readjusting k."""
        length = len(s)
        self.b = self.b[:self.j+1] + s + self.b[self.j+length+1:]
        self.k = self.j + length

    def r(self, s):
        """r(s) is used further down."""
        if self.m() > 0:
            self.setto(s)

    def step1ab(self):
        """step1ab() gets rid of plurals and -ed or -ing. e.g.

           caresses  ->  caress
           ponies    ->  poni
           ties      ->  ti
           caress    ->  caress
           cats      ->  cat

           feed      ->  feed
           agreed    ->  agree
           disabled  ->  disable

           matting   ->  mat
           mating    ->  mate
           meeting   ->  meet
           milling   ->  mill
           messing   ->  mess

           meetings  ->  meet
        """
        if self.b[self.k] == 's':
            if self.ends("sses"):
                self.k = self.k - 2
            elif self.ends("ies"):
                self.setto("i")
            elif self.b[self.k - 1] != 's':
                self.k = self.k - 1
        if self.ends("eed"):
            if self.m() > 0:
                self.k = self.k - 1
        elif (self.ends("ed") or self.ends("ing")) and self.vowelinstem():
            self.k = self.j
            if self.ends("at"):   self.setto("ate")
            elif self.ends("bl"): self.setto("ble")
            elif self.ends("iz"): self.setto("ize")
            elif self.doublec(self.k):
                self.k = self.k - 1
                ch = self.b[self.k]
                if ch == 'l' or ch == 's' or ch == 'z':
                    self.k = self.k + 1
            elif (self.m() == 1 and self.cvc(self.k)):
                self.setto("e")

    def step1c(self):
        """step1c() turns terminal y to i when there is another vowel in the stem."""
        if (self.ends("y") and self.vowelinstem()):
            self.b = self.b[:self.k] + 'i' + self.b[self.k+1:]

    def step2(self):
        """step2() maps double suffices to single ones.
        so -ization ( = -ize plus -ation) maps to -ize etc. note that the
        string before the suffix must give m() > 0.
        """
        if self.b[self.k - 1] == 'a':
            if self.ends("ational"):   self.r("ate")
            elif self.ends("tional"):  self.r("tion")
        elif self.b[self.k - 1] == 'c':
            if self.ends("enci"):      self.r("ence")
            elif self.ends("anci"):    self.r("ance")
        elif self.b[self.k - 1] == 'e':
            if self.ends("izer"):      self.r("ize")
        elif self.b[self.k - 1] == 'l':
            if self.ends("bli"):       self.r("ble") # --DEPARTURE--
            # To match the published algorithm, replace this phrase with
            #   if self.ends("abli"):      self.r("able")
            elif self.ends("alli"):    self.r("al")
            elif self.ends("entli"):   self.r("ent")
            elif self.ends("eli"):     self.r("e")
            elif self.ends("ousli"):   self.r("ous")
        elif self.b[self.k - 1] == 'o':
            if self.ends("ization"):   self.r("ize")
            elif self.ends("ation"):   self.r("ate")
            elif self.ends("ator"):    self.r("ate")
        elif self.b[self.k - 1] == 's':
            if self.ends("alism"):     self.r("al")
            elif self.ends("iveness"): self.r("ive")
            elif self.ends("fulness"): self.r("ful")
            elif self.ends("ousness"): self.r("ous")
        elif self.b[self.k - 1] == 't':
            if self.ends("aliti"):     self.r("al")
            elif self.ends("iviti"):   self.r("ive")
            elif self.ends("biliti"):  self.r("ble")
        elif self.b[self.k - 1] == 'g': # --DEPARTURE--
            if self.ends("logi"):      self.r("log")
        # To match the published algorithm, delete this phrase

    def step3(self):
        """step3() dels with -ic-, -full, -ness etc. similar strategy to step2."""
        if self.b[self.k] == 'e':
            if self.ends("icate"):     self.r("ic")
            elif self.ends("ative"):   self.r("")
            elif self.ends("alize"):   self.r("al")
        elif self.b[self.k] == 'i':
            if self.ends("iciti"):     self.r("ic")
        elif self.b[self.k] == 'l':
            if self.ends("ical"):      self.r("ic")
            elif self.ends("ful"):     self.r("")
        elif self.b[self.k] == 's':
            if self.ends("ness"):      self.r("")

    def step4(self):
        """step4() takes off -ant, -ence etc., in context <c>vcvc<v>."""
        if self.b[self.k - 1] == 'a':
            if self.ends("al"): pass
            else: return
        elif self.b[self.k - 1] == 'c':
            if self.ends("ance"): pass
            elif self.ends("ence"): pass
            else: return
        elif self.b[self.k - 1] == 'e':
            if self.ends("er"): pass
            else: return
        elif self.b[self.k - 1] == 'i':
            if self.ends("ic"): pass
            else: return
        elif self.b[self.k - 1] == 'l':
            if self.ends("able"): pass
            elif self.ends("ible"): pass
            else: return
        elif self.b[self.k - 1] == 'n':
            if self.ends("ant"): pass
            elif self.ends("ement"): pass
            elif self.ends("ment"): pass
            elif self.ends("ent"): pass
            else: return
        elif self.b[self.k - 1] == 'o':
            if self.ends("ion") and (self.b[self.j] == 's' or self.b[self.j] == 't'): pass
            elif self.ends("ou"): pass
            # takes care of -ous
            else: return
        elif self.b[self.k - 1] == 's':
            if self.ends("ism"): pass
            else: return
        elif self.b[self.k - 1] == 't':
            if self.ends("ate"): pass
            elif self.ends("iti"): pass
            else: return
        elif self.b[self.k - 1] == 'u':
            if self.ends("ous"): pass
            else: return
        elif self.b[self.k - 1] == 'v':
            if self.ends("ive"): pass
            else: return
        elif self.b[self.k - 1] == 'z':
            if self.ends("ize"): pass
            else: return
        else:
            return
        if self.m() > 1:
            self.k = self.j

    def step5(self):
        """step5() removes a final -e if m() > 1, and changes -ll to -l if
        m() > 1.
        """
        self.j = self.k
        if self.b[self.k] == 'e':
            a = self.m()
            if a > 1 or (a == 1 and not self.cvc(self.k-1)):
                self.k = self.k - 1
        if self.b[self.k] == 'l' and self.doublec(self.k) and self.m() > 1:
            self.k = self.k -1

    def stem(self, p, i, j):
        """In stem(p,i,j), p is a char pointer, and the string to be stemmed
        is from p[i] to p[j] inclusive. Typically i is zero and j is the
        offset to the last character of a string, (p[j+1] == '\0'). The
        stemmer adjusts the characters p[i] ... p[j] and returns the new
        end-point of the string, k. Stemming never increases word length, so
        i <= k <= j. To turn the stemmer into a module, declare 'stem' as
        extern, and delete the remainder of this file.
        """
        # copy the parameters into statics
        self.b = p
        self.k = j
        self.k0 = i
        if self.k <= self.k0 + 1:
            return self.b # --DEPARTURE--

        # With this line, strings of length 1 or 2 don't go through the
        # stemming process, although no mention is made of this in the
        # published algorithm. Remove the line to match the published
        # algorithm.

        self.step1ab()
        self.step1c()
        self.step2()
        self.step3()
        self.step4()
        self.step5()
        return self.b[self.k0:self.k+1]

class GOOboXmlHandler (ContentHandler):
    """ This class handles parsing the OBO XML file from the Gene Ontology Consortium and populate
        a GOGraph object passed to the constructor of this class.  The handler only use the nodes from
        the namespace (a branch of the GO) specificed in the GOGraph object.  Currently, only the edges
        that reflect IS_A relationship between a pair of GO terms are parsed and added into the GOGraph

        Modification by AH: 
        1. also add PART_OF relationships into the graph
        2. return a list of synonyms

        from gographer.GOOboXmlHandler.py
    """
    ## Constructor.
    # @param    goGraph A reference to a GOGraph object to which the parsed GO nodes and edges
    #                   will be added.
    def __init__(self, goGraph):
        self.namespace = goGraph.getNameSpace()
        self.graph = goGraph
        self.inTypedef = False
        self.goNode = GONode()
        self.obsolete = False
        self.relationship_type = None
        self.synonyms = {}
        self.obsoletes = set()

    def startElement(self, name, attributes):
        self.cdata = ""

        if name == "term":
            self.goid = None
            self.name = None
            self.description = None
            self.obsolete = False            
            self.elementNamespace = None
            self.parents = []
            self.goNode = GONode()
            
        elif name == "typedef":
            self.inTypedef = True

    def endElement(self, name):

        # print name, self.obsolete, self.namespace #, self.elementNamespace

        if name == "id":
            self.goid = self.cdata.strip()
            self.goNode.setGOID(self.goid)

        elif name == "namespace":
            self.elementNamespace = self.cdata.strip()
            self.goNode.setNamespace(self.elementNamespace)

        elif name == "is_a" and not self.inTypedef:
            self.parents.append(self.cdata.strip())
            self.goNode.setParents(self.parents)

        elif name == "type":
            self.relationship_type = self.cdata.strip()

        elif name == "relationship" and not self.inTypedef:
            # add relationships into graph - this includes 'part_of' but excludes
            # 'negatively_regulates', 'positively_regulates' and 'regulates' and others
            if self.relationship_type == "part_of":
                self.parents.append(self.cdata.strip())
                self.goNode.setParents(self.parents)
            
        elif name == "is_obsolete" and self.cdata.strip() == "1":
            self.obsolete = True
            self.goNode.setObsolete(self.obsolete)

        elif name == "typedef":
            self.inTypedef = False

        elif name == "alt_id":
            self.synonyms[self.cdata.strip()] = self.goid

        elif name == "term" and not self.obsolete and self.namespace == self.elementNamespace:
            
            '''# not sure what the follow trying to do
            if not self.elementNamespace in self.namespaces.keys():
                self.namespaces[self.elementNamespace] = []

            self.namespaces[self.elementNamespace].append(self.goid)
            '''
            
            self.graph.add_node(self.goid, data=self.goNode)
            for parent in self.parents:
                '''
                currently the only edges being added are to the children of the current node
                '''
                if not parent in self.graph:
                    self.graph.add_node(parent, data=GONode(parent))
                self.graph.add_edge(parent, self.goid, relationship="parent_of")

        elif name == "name" and not self.obsolete and not self.goid == None and self.goNode.name == None: # and not self.name.has_key(self.goid):
            self.goNode.setName(self.cdata.strip())

        elif name == "defstr" and  not self.obsolete and not self.goid == None: # and not self.name.has_key(self.goid):
            self.goNode.setDescription(self.cdata.strip())

        elif name == "term" and self.obsolete and self.namespace == self.elementNamespace:
            self.obsoletes.add( self.goid )

    def characters(self, data):
        self.cdata += data

class GONode():
    '''from gographer.GONode.py'''
    def __init__ (self, goid=None, namespace=None, parents=None, obsolete=False,
                  name=None, description=None, genes=set(), propGenes=set(),
                  pmids=set(), propPmids=set(), wordVector=dict(), descendantCount=None,
                  mergedGenes=set(), mergedPmids=set(), mergedCount=0, infoLoss=0):
        self.goid = goid
        self.namespace = namespace
        self.parents = parents
        self.obsolete = obsolete
        self.name = name
        self.description = description
        self.genes = genes
        self.propGenes = propGenes
        self.pmids = pmids
        self.propPmids = propPmids
        self.wordVector = wordVector
        self.descendantCount = descendantCount
        self.mergedGenes = mergedGenes
        self.mergedPmids = mergedPmids
        self.mergedCount = mergedCount
        self.infoLoss = infoLoss


    ##Set the GO ID of the node
    # @param    goid    The GO ID that will be assigned to the node
    def setGOID (self, goid):
        self.goid = goid

    ##Returns the GO ID of the node
    def getGOID (self):
        return self.goid

    ##Set the namespace of the node
    # @param    namespace   The namespace that will be assigned to the node 
    def setNamespace (self, namespace):
        self.namespace = namespace

    ##Returns the namespace of the node
    def getNamespace (self):
        return self.namespace

    ##Set the parents of the node
    # @param    parents  The parents that will be assigned to the node
    def setParents (self, parents):
        self.parents = parents
        
    ##Returns the parents of the node
    def getParents (self):
        return self.parents

    ##Set the obsolete status of the node
    # @param    obsolete    The obsolete status that will be assigned to the node
    def setObsolete (self, obsolete):
        self.obsolete = obsolete

    ##Returns the obsolete status of the node (whether or not the term is obsolete)
    def getObsolete (self):
        return self.obsolete

    ##Set the name of the node
    # @param    name    The name that will be assigned to the node
    def setName (self, name):
        self.name = name

    ##Returns the name of the node
    def getName (self):
        return self.name
    
    ##Set the description of the node
    # @param    description The description that will be assigned to the node
    def setDescription (self, description):
        self.description = description

    ##Returns the description of the node
    def getDescription (self):
        return self.description

    ##Set the PubMed IDs associated with the node
    # @param    pmids   The set containing the PubMed IDs that are associated with the node.
    #                   The PubMed IDs are in tuples where it's the PubMed ID number followed by the qualifier.
    def setPMIDs (self, pmids):
        self.pmids = pmids

    ##Returns the PubMed IDs associated with the node
    def getPMIDs (self):
        return self.pmids

    ##Adds list containing one or more PubMed ID tuples to the existing set of PubMed IDs
    # @param    pmid    A list where each entry is a PubMed ID tuple, where it's the PubMed ID number followed by the qualifier
    def addPMIDs(self, pmid):
        self.pmids = self.pmids.union(pmid)

    ##Set the propagated PubMed IDs associated with the node or its descendants
    # @param    propPmids   The set containing the propagated PubMed IDs that are associated with the node or its descendants.
    #                   The PubMed IDs are in tuples where it's the PubMed ID number followed by the qualifier.
    def setPropagatedPMIDs (self, propPmids):
        self.propPmids = propPmids

    ##Returns the propagated PubMed IDs associated with the node or its descendants
    def getPropagatedPMIDs (self):
        return self.propPmids

    ##Adds list containing PubMed ID tuples to the existing set of propagated PubMed IDs
    # @param    pmids    A list where each entry is a PubMed ID tuple, where it's the PubMed ID number followed by the qualifier
    def addPropagatedPMIDs(self, pmids):
        self.propPmids = self.propPmids.union(pmids)

    ##Set the genes associated with the node
    # @param    genes   The set containing the genes that are associated with the node.
    #                   The genes are in tuples where it's the gene ID followed by the qualifier.
    def setGenes (self, genes):
        self.genes = genes

    ##Returns the genes associated with the node
    def getGenes (self):
        return self.genes

    ##Adds list containing one or more gene tuples to the existing set of genes
    # @param    genes    A list where each entry is a gene tuple, where it's the gene ID followed by the qualifier
    def addGenes(self, genes):
        self.genes = self.genes.union(genes)

    ##Set the propagated genes associated with the node or its descendants
    # @param    propGenes   The set containing the propagated PubMed IDs that are associated with the node or its descendants.
    #                   The PubMed IDs are in tuples where it's the PubMed ID number followed by the qualifier.
    def setPropagatedGenes (self, propGenes):
        self.propGenes = propGenes

    ##Returns the propagated PubMed IDs associated with the node or its descendants
    def getPropagatedGenes (self):
        return self.propGenes

    ##Adds list containing PubMed ID tuples to the existing set of propagated PubMed IDs
    # @param    pmids    A list where each entry is a PubMed ID tuple, where it's the PubMed ID number followed by the qualifier
    def addPropagatedGenes(self, genes):
        self.propGenes = self.propGenes.union(genes)

    ##Calculates the word vector and stores this information
    # @param    corpus  The corpus that contains the information on the PubMed article
    # @param    pmids   A list of PubMed ID tuples to be added to the word vector, where it's the PubMed ID number followed by the qualifier
    #                   The propagated PubMed IDs will be used if none is given
    # @param    tokenizer   The tokenizer function that will be used on the text, a simple tokenizer is used if none is given.
    #                       Should take a string as an input, and outputs a string with words that are lower case and separated by a space
    # @param    stemmer The stemmer function that will be used to stem the words, the porter stemmer is used if none is given
    #                   Takes a word as an input and reports a stemmed word as an output.
    # @param    stopwords   A list of stop words that will not be included in the word vector, either as a list or a StopwordList.
    #                       An empty list is used if no stop word list is given.
    def calculateWordVector(self, corpus, tokenizer=Tokenizer().tokenize_word,
                            stemmer=PorterStemmer().stem, stopwords=[], pmids=None):
        wordVector = {}

        if not corpus:
            print "Not corpus is given, so a word vector can not be calculated."

        else:
            if not pmids:
                pmids = self.getPropagatedPMIDs()

            for pmid, qualifier in pmids:
                #Checks to make sure the PubMed article is in the corpus
                #and the qualifier does not contain the word 'NOT'
                if pmid in corpus.docs and "NOT" not in qualifier:
                    #Adds each word to the word vector
                    pmidWordVector = corpus.docs[pmid].getWordVector(tokenizer, stemmer, stopwords)
                    for word in pmidWordVector:
                        #if word not in stopwords:
                        if word in wordVector:
                            wordVector[word] += pmidWordVector[word]
                        else:
                            wordVector[word] = pmidWordVector[word]
        self.wordVector = wordVector
    
    ##Returns the word vector for the node calculated using propagated PubMed IDs
    # @param    corpus  The corpus that contains the information on the PubMed article
    def getWordVector(self, corpus=None, stopwords=[]):
        #Calculates the word vector if the current word vector is empty
        if len(self.wordVector) == 0:
            if len(self.getPropagatedPMIDs()) > 0:
                self.calculateWordVector(corpus, stopwords=stopwords)
        return self.wordVector

    ##Returns the number of descendants of the current node
    def getDescendantCount(self):
        return self.descendantCount

    
    ##Set the genes associated with the nodes that have been merged into the current node
    # @param    mergedGenes   The set containing the genes that are associated with the merged nodes.
    #                   The genes are in tuples where it's the gene ID followed by the qualifier.
    def setMergedGenes (self, mergedGenes):
        self.mergedGenes = mergedGenes

    ##Returns the genes associated with the nodes that have been merged into the current node
    def getMergedGenes (self):
        return self.mergedGenes

    ##Adds list containing one or more gene tuples to the existing set of merged genes
    # @param    mergedGenes    A list where each entry is a gene tuple, where it's the gene ID followed by the qualifier
    def addMergedGenes(self, mergedGenes):
        self.mergedGenes = self.mergedGenes.union(mergedGenes)


    ##Set the PubMed IDs associated with the nodes that have been merged into the current node
    # @param    mergedPmids   The set containing the PubMed IDs that are associated with the merged nodes.
    #                   The PubMed IDs are in tuples where it's the PubMed ID number followed by the qualifier.
    def setMergedPMIDs (self, mergedPmids):
        self.mergedPmids = mergedPmids

    ##Returns the PubMed IDs associated with the nodes that have been merged into the current node
    def getMergedPMIDs (self):
        return self.mergedPmids

    ##Adds list containing one or more PMID tuples to the existing set of merged PMIDs
    # @param    mergedPmids    A list where each entry is a PubMed ID tuple, where it's the PubMed ID followed by the qualifier
    def addMergedPMIDs(self, mergedPmids):
        self.mergedPmids = self.mergedPmids.union(mergedPmids)

    ##Set the count of the number of nodes that have been merged into the current node
    # @param    mergedCount   The integer count of the number of nodes that have been merged into the current node
    def setMergedCount (self, mergedCount):
        self.mergedCount = mergedCount

    ##Returns the PubMed IDs associated with the nodes that have been merged into the current node
    def getMergedCount (self):
        return self.mergedCount

    ##Adds list containing one or more PMID tuples to the existing set of merged PMIDs
    # @param    mergedPmids    A list where each entry is a PubMed ID tuple, where it's the PubMed ID followed by the qualifier
    def addMergedCount(self, mergedCount=1):
        self.mergedCount += mergedCount

    ##Set the PubMed IDs associated with the nodes that have been merged into the current node
    # @param    infoLoss   The set containing the PubMed IDs that are associated with the merged nodes.
    #                   The PubMed IDs are in tuples where it's the PubMed ID number followed by the qualifier.
    def setInfoLoss (self, infoLoss):
        self.infoLoss = infoLoss

    ##Returns the PubMed IDs associated with the nodes that have been merged into the current node
    def getInfoLoss (self):
        return self.infoLoss

    ##Adds list containing one or more PMID tuples to the existing set of merged PMIDs
    # @param    infoLoss    A list where each entry is a PubMed ID tuple, where it's the PubMed ID followed by the qualifier
    def addInfoLoss(self, infoLoss):
        self.infoLoss += infoLoss
        
class GOGraph(networkx.DiGraph):
    '''from gographer.GOGraph.py'''
    ## Constructor.
    # @param    namespace     The branch of the GO ontology stored by this graph
    # @param    XMLFileName   The file contains the GO definition in
    #                         the format of the OBO in XML
    def __init__(self, namespace=None, GOOboXmlFileName=None):        
        DiGraph.__init__(self)

        self.namespace = namespace

        if GOOboXmlFileName != None:
            self.parseOboXml(GOOboXmlFileName)
        
    ## Parses the given OBO XML file and creates a GOGraph
    # @param    GOOboXMLFileName    The name of the OBO XML file to be parsed
    def parseOboXml(self, GOOboXmlFileName):
        parser = make_parser()
        handler = GOOboXmlHandler(self)
        parser.setContentHandler(handler)
        f = IOTools.openFile(GOOboXmlFileName, 'r')
        parser.parse(f)
        f.close()
        self.synonyms = handler.synonyms
        self.obsoletes = handler.obsoletes

    ## Returns the minimum depth of the node
    # @param    goid    The GO ID of the node whose depth will be found and returned
    def getLevel(self, goid):
        parents = self.predecessors(goid)
        if len(parents) == 0:
            return 0
        
        found = False
        level = 1
        while not found:
            nextDepth = list()
            for i in parents:
                pred = self.predecessors(i)
                if len(pred) == 0:
                    found = True
                    return level
                else:
                    nextDepth = list(set(nextDepth) | set(pred))
            parents = nextDepth
            level = level + 1


    ## Returns the namespace of the graph
    def getNameSpace(self):
        return self.namespace;

    ## Save self using pickle protocol.
    # @param filename The filename of the pickle to use
    def savePickle(self, filename="gograph.pickle"):
        try:
            f = open(filename, 'wb')
            cPickle.dump(self, f, protocol = cPickle.HIGHEST_PROTOCOL)
            f.close()
        except:
            print "Could not pickle graph"

    ## Load a pickle from the filesystem
    # @param    filename    The location of the pickle file to load
    @classmethod
    def loadPickle (klass, filename="gograph.pickle"):
        try:
            f = open(filename, 'rb')
            g = cPickle.load(f)
            return g
        except:
            print "Could not load pickle"
            return

    ## Get the description of a node
    # @param    goid   The GO ID of the node to get the description of
    def getNodeDescription(self, goid):
        if goid in self.nodes():
            return self.node[goid]['data'].getDescription()
        else:
            print 'Invalid goid'
            raise

    ## Set the description of a node
    # @param    goid   The GO ID of the node to get the description of
    # @param    descrip The description that will be assigned to the node
    def __setNodeDescription(self, goid, descrip):
        if goid in self.nodes():
            self.node[goid]['data'].setDescription(descrip)
        else:
            print 'Invalid goid'
            raise

    ## Get the description of a node
    # @param    goid   The GO ID of the node to get the description of
    def getNodeNamespace(self, goid):
        if goid in self.nodes():
            return self.node[goid]['data'].getNamespace()
        else:
            print 'Invalid goid'
            raise

    ## Set the description of a node
    # @param    goid   The GO ID of the node to get the description of
    # @param    namespace The namespace that will be assigned to the node
    def __setNodeNamespace(self, goid, namespace):
        if goid in self.nodes():
            self.node[goid]['namespace'] = namespace
        else:
            print 'Invalid goid'
            raise

    ## Calculates and stores the number of descendants each node in the graph has
    def calcDescendantCount(self):
        nodes = dict()
        sortedNodes = topological_sort(self)
        sortedNodes.reverse()
        for node in sortedNodes:
            if node not in nodes:
                ids = set()
            else:
                ids = nodes[node]

            self.node[node]['data'].descendantCount = len(ids)
            ids = ids.union([node])

            for parent in self.predecessors(node):
                if parent not in nodes:
                    nodes[parent] = ids
                else:
                    nodes[parent] = nodes[parent].union(ids)

    ## Return the number of descendants of the given node. Calculates the count if it had not already been done so
    # @param    goid    The GO ID of the node to return the descendant count of
    def getDescendantCount(self, goid):
        if goid in self:
            count = self.node[goid]['data'].getDescendantCount()
            if not count:
                self.calcDescendantCount()
                return self.node[goid]['data'].getDescendantCount()
            else:
                return count



def getAncestors( g, node ):
    p = g.predecessors( node )
    ancestors = set(p)
    work = p[:]
    while work:
        x = work.pop()
        p = g.predecessors( x )
        work.extend( [ x for x in p if x not in ancestors ])
        ancestors.update(p)

    return ancestors

def readPValues( infile, synonyms = {} ):
    '''read pvalues and optionally fold changes from infile.'''
    term2pvalue, term2log2fold = {}, collections.defaultdict( float )

    def _parse2(data):
        goid = synonyms.get( data[0], data[0] )
        term2pvalue[goid] = float(data[1])

    def _parse3(data):
        goid = synonyms.get( data[0], data[0] )
        term2pvalue[goid] = float(data[1])
        term2log2fold[goid] = math.log( float(data[2]), 2 )

    def _parseGO(data):
        '''parse output from GO.py'''
        if len(data) < 14: return
        goid = synonyms.get( data[11], data[11] )
        term2pvalue[goid] = float( data[8] )
        term2log2fold[goid] = math.log( float(data[7])+0.0000001, 2 )

    parser = None
    for line in infile:
        if line.startswith("code"): 
            parser = _parseGO
            continue
        elif line.startswith("goid"):
            data = line[:-1].split("\t")
            if len(data) == 2:
                parser = _parse2
            else:
                parser = _parse3
            continue
        if parser == None:
            raise ValueError("don't know how to parse this file at line %s" % line)
        
        parser( line[:-1].split() )

    return term2pvalue, term2log2fold

def readSimrelMatrix( infile ):
    '''read simrel matrix from file.'''
    matrix, row_headers, col_headers = IOTools.readMatrix( infile )
    assert row_headers == col_headers
    return matrix, row_headers

def readTermFrequencies( infile ):
    '''read term frequencies and counts from file.'''
    counts, p = {}, {}
    for line in infile:
        if line.startswith("goid"):continue
        goid, f, c = line[:-1].split("\t")
        p[goid] = float(f)
        counts[goid] = set(c.split(";"))

    return counts, p

def readAncestors( infile ):
    '''read term frequencies and counts from file.'''
    ancestors = {}
    for line in infile:
        if line.startswith("goid"):continue
        goid, anc = line[:-1].split("\t")
        ancestors[goid] = set(anc.split(";"))
    return ancestors

def computeAncestors( graph, nodes ):
    '''build dictionary with node ancestors in graph.'''

    iteration, total = 0, len(nodes)

    E.info("computing ancestral information for %i nodes" % len(nodes))

    ancestors = {}
    for node in nodes:
        a = getAncestors( graph, node )
        # only take nodes with annotations
        a = set([ x for x in a if x in nodes ])
        # add self
        # a.add( node )
        ancestors[node] = a 
        iteration += 1
        if iteration % 500 == 0: 
            E.debug( "total = %i (%5.2f%%)" % (iteration, 100.0 * iteration / total) )
    return ancestors

def getRoot( graph ):
    '''return root of graph.'''
    root = graph.nodes()[0]
    p = graph.predecessors( root )
    while p != []:
        root = p[0]
        p = graph.predecessors( root )
    return root

def buildFilename( ontology, section, suffix ):
    return E.getOutputFile( "%s.%s.%s" % (ontology,section,suffix) )

def computeTermFrequencies( graph, go2genes ):

    nodes = graph.nodes()
    
    root = getRoot( graph )

    ############################################################
    # compute annotation counts - number of times a term
    # occurs in the background database
    E.info("computing annotation counts" )
    counts = collections.defaultdict( set )

    ############################################################
    # compute term frequencies p[term]
    # eq. 1 in Schlicker et al. (2006)
    E.info("computing term frequencies" )
    # compute counts
    for node in traversal.dfs_postorder( graph, source = root ):
    # older networkx:
    # for node in traversal.dfs_postorder_nodes( graph, source = root ):
        counts[node] = set(go2genes[node])
        for x in graph.successors( node ):
            counts[node].update( counts[x] )

    # normalize
    rf = float(len(counts[root]))
    p = {}
    for term, v in counts.iteritems():
        p[term] = len(v) / rf
        
    E.info( "computed %i frequencies" % len(p) )

    # debugging code - check that frequencies are increasing
    # monotonically.
    for node in nodes:
        pre = graph.predecessors( node )
        while pre != []:
            assert p[pre[0]] >= p[node]
            node = pre[0]
            pre = graph.predecessors( node )

    return counts, p

def buildSimrelMatrix( go2genes, go2info, ancestors, p ):
    '''build simrel matrix from an ontology DAG in *graph*
    and a mapping of go terms to genes.

    This method follows the procedure and syntax
    according to Schlicker et al. (2006) 
    BMC Bioinformatics 7:302
    '''

    ############################################################
    # Build symmetric similarity matrix between terms

    # only work on nodes with annotations
    lp = {}
    for term, v in p.iteritems():
        # ignore p = 1.0 as these will be downweighted to 
        # 0 (1.0 - p[c])
        if v > 0 and v < 1.0: lp[term] = math.log( v )

    test_nodes = lp.keys()
    node2index = dict( [ (y,x) for x,y in enumerate( test_nodes ) ] )

    # compute dictionary of ancestors for
    # each node. A simple dfs does not work as the
    # graph is not a tree.
    E.info("building %i x %i simrel matrix" % (len(test_nodes), len(test_nodes)))
    matrix = numpy.zeros( (len(test_nodes), len(test_nodes)) )
    iteration = 0

    total = len(test_nodes) * (len(test_nodes) - 1) / 2

    for c1, c2 in itertools.combinations( test_nodes, 2 ):
        iteration += 1
        if iteration % 1000000 == 0: 
            E.info( "total = %i (%5.2f%%)" % (iteration, 100.0 * iteration / total) )

        s1 = ancestors[c1]
        s2 = ancestors[c2]
        
        # print "c1=", c1, p[c1], lp[c1]
        # outputList( s1, go2info, p, lp )

        # print "c2=", c2, p[c2], lp[c2]
        # outputList( s2, go2info, p, lp )

        common = s1.intersection( s2 )

        i,j = node2index[c1], node2index[c2]

        f = 2.0 / (lp[c1] + lp[c2])
        # add pseudocount to avoid division by zero 
        vals = [ lp[c] * (1.0 - p[c] ) for c in common if c in lp ]
        if len(vals) > 0:
            # use min as vals are multiplied by negative factor
            val = f * min(vals)
            assert 0 <= val <= 1
            matrix[i][j] = val
            matrix[j][i] = val

    return matrix, test_nodes

def clusterSimrelMatrix( matrix, terms, ancestors, p, counts, term2pvalue, go2info, options):
    '''apply revigo clustering scheme to *matrix*.

    *ancestors*, *p*, and *counts* contain auxiliary data

    ancestors: dictionary mapping a term to all its ancestors
    p: dictionary mapping a term to its annotation frequency
    counts: dictionary mapping a term to all the genes it is annotated with
    go2info: dictionary mapping a term to its description
    term2pvalue: dictionary mapping a term to its p-value

    returns (matrix, terms)
    '''

    # cluster or load clustered simrel matrix
    E.info("working on simrel matrix with %i terms" % len(terms))
    term2index = dict( [ (y,x) for x,y in enumerate( terms ) ] )

    # shrink matrix to only those terms that are in result set
    terms = [ x for x in terms if x in term2pvalue ]

    E.info("after removing terms with no p-value: %i terms" % len(terms))
    if len(terms) == 0:
        E.warn( "no terms given after p-value filtering" )
        return None, [], []

    matrix = matrix[ numpy.array( [ term2index[x] for x in terms ]) ]
    matrix = matrix[ :, numpy.array( [ term2index[x] for x in terms ]) ]
    term2index = dict( [ (y,x) for x,y in enumerate( terms ) ] )

    # cluster according to revigo
    clusters = dict( [ (x, [x]) for x in terms ] )
    nrows = len(terms)

    counter = E.Counter()

    # row/column indices of members
    member_indices = []

    while 1:
        # find most similar pair of GO terms ti and tj with similarity simrel
        index = numpy.argmax( matrix )
        i = index // nrows
        j = index % nrows

        simrel = matrix[i][j]
        assert simrel == matrix.max()            

        # stop procedure if similarity is below threshold
        if simrel == 0:
            E.debug("similarity reached 0 - done" )
            break

        if simrel < options.max_simrel_threshold: 
            E.debug("similarity %f < %f - done" % (simrel, options.max_simrel_threshold))
            break

        nclusters = len(clusters)
        if nclusters <= options.min_clusters_threshold:
            E.debug("%i clusters reached at similarity %f - done" % (nclusters, simrel) )
            break

        t1 = terms[i]
        t2 = terms[j]
        E.debug( "clustering at similarity %f (%i,%i)" % (simrel, i,j))

        p1 = term2pvalue[t1]
        p2 = term2pvalue[t2]

        E.debug( "1. %s (%s): pvalue=%f, freq=%f" % (t1, go2info.get(t1, "na"), p1, p[t1] * 100.0))
        E.debug( "2. %s (%s): pvalue=%f, freq=%f" % (t2, go2info.get(t2, "na"), p2, p[t2] * 100.0))

        counter.tested += 1

        if p[t1] > options.min_frequency:
            # remove term with frequency > 5%
            counter.min_frequency += 1
            rep, mem = t2, t1

        elif p[t2] > options.min_frequency:
            # remove term with frequency > 5%
            counter.min_frequency += 1
            rep, mem = t1, t2

        elif term2pvalue[t1] < term2pvalue[t2]:
            # remove term with higher pvalue
            counter.min_pvalue += 1
            rep, mem = t1, t2

        elif term2pvalue[t2] < term2pvalue[t1]:
            # remove term with higher pvalue
            counter.min_pvalue += 1
            rep, mem = t2, t1

        elif t1 in ancestors[t2]:
            # t1 is parent of t2 - reject child term t2
            ovl = float(len(counts[t1].intersection( counts[t2] )))
            if (ovl / len(counts[t1])) >= options.min_child_threshold:
                # exception: if parent is > 75% child, keep child
                counter.ancestor_child += 1
                rep, mem = t2, t1
            else:
                counter.ancestor_parent += 1
                rep, mem = t1, t2

        elif t2 in ancestors[t1]:
            # t2 is parent of t1 - reject child term t1
            ovl = float(len(counts[t1].intersection( counts[t2] )))
            if (ovl / len(counts[t2])) >= options.min_child_threshold:
                # exception: if parent is > 75% child, keep child
                counter.ancestor_child += 1
                rep, mem = t1, t2
            else:
                counter.ancestor_parent += 1
                rep, mem = t2, t1

        else:
            # keep random term
            counter.random += 1
            rep, mem = min(t1,t2), max(t1,t2)

        # merge clusters
        clusters[rep].extend( clusters[mem] )
        del clusters[mem]
        # set values of mem to 0
        if mem == t1: 
            matrix[i,:] = 0
            matrix[:,i] = 0
            member_indices.append( i )
        else:
            matrix[j,:] = 0
            matrix[:,j] = 0
            member_indices.append( j )

    E.info( "%s" % str(counter))
    E.info( "after clustering: %i clusters" % len(clusters) )

    # shrink matrix
    member_indices = set( member_indices )
    rep_indices = [ x for x in xrange(nrows) if x not in member_indices ]
    terms = [ terms[x] for x in rep_indices ]
    rows = len(terms)
    matrix = matrix[ numpy.array( [ x for x in rep_indices ]) ]
    matrix = matrix[ :, numpy.array( [ x for x in rep_indices ]) ]

    return matrix, terms, clusters

@E.cachedfunction
def readGOGraph( filename_obo, namespace ):
    E.info( "reading GO graph from %s: namespace =%s" % (filename_obo, namespace))
    graph = GOGraph( namespace = namespace, GOOboXmlFileName=filename_obo )
    return graph

def main( argv ):

    parser = E.OptionParser( version = "%prog version: $Id$", 
                                    usage = globals()["__doc__"])

    parser.add_option("-o", "--filename-ontology", dest="filename_obo", type="string",
                      help="filename with ontology information in obo-xml format. "
                      " The latest version can always be"
                      " retrieved at `wget http://archive.geneontology.org/latest-termdb/go_daily-termdb.obo-xml.gz`."
                      " [default=%default]." )

    parser.add_option("-g", "--filename-go", dest="filename_go", type="string",
                      help="filename with gene to go assignments "
                      " This is a tab-separated file with the columns <namespace> <gene_id> <goid> <description> <evidence>." 
                      " The evidence column is currently ignored. "
                      " [default=%default]." )

    parser.add_option("-p", "--filename-pvalues", dest="filename_pvalues", type="string",
                      help="filename with pvalues for each GO term "
                      " This is a tab-separated file with the columns <goid> <pvalue>."
                      " If a third column is present it will be interpreted as fold-change. "
                      " [default=%default]." )

    parser.add_option( "--ontology", dest="ontology", type="choice", action="append",
                       choices=("biol_process","cell_location","mol_function", "all" ),
                       help="ontologies to analyze. Ontologies are tested separately."
                       " 'all' will test all ontologies. "
                       " [default=%default]." )

    parser.add_option( "--max-similarity", dest="max_simrel_threshold", type="float",
                      help="cluster until simrel threshold is achieved [default=%default]." )

    parser.add_option( "--min-clusters", dest="min_clusters_threshold", type="float",
                      help="cluster only until at most # clusters remain [default=%default]." )

    parser.add_option( "--palette", dest="palette", type="choice",
                       choices=("rainbow", "gray", "blue-white-red",
                                "autumn", "bone", "cool", "copper", "flag", "gray", 
                                "hot", "hsv", "jet", "pink", "prism",
                                "spring", "summer", "winter", "spectral",
                                "RdBu", "RdGy", "BrBG", "BuGn", "Blues", "Greens", "Reds", "Oranges", "Greys" ),
                       help="colour palette [default=%default]")
    

    parser.add_option( "--reverse-palette", dest="reverse_palette", action="store_true",
                      help="reverse colour palette [default=%default]." )

    parser.set_defaults( filename_obo = 'go_daily-termdb.obo-xml.gz',
                         filename_go = "go.tsv.gz",
                         #filename_pvalues = "/ifs/projects/proj008/report/genelists.go.dir/rela_upstream-changed.biol_process.results", 
                         filename_pvalues = None,
                         filename_bg = None, 
                         filename_matrix = None,
                         filename_frequencies = None,
                         ontology = [],
                         # 0.9, 0.7, 0.5 and 0.4
                         # 0.4 very small clusters
                         max_simrel_threshold = 0.7,
                         min_frequency = 0.05,
                         min_pvalue = 0.05,
                         min_child_threshold = 0.75,
                         min_clusters_threshold = 0,
                         palette = "RdBu",
                         reverse_palette = False,
                         )

    options, args = E.Start( parser, argv, add_output_options = True )

    map_ontology2namespace = {
        'biol_process' :"biological_process",
        'cell_location': 'cellular_component',
        'mol_function': 'molecular_function' }

    if "all" in options.ontology:
        options.ontology = map_ontology2namespace.keys()

    #########################################
    if options.filename_pvalues == None:
        E.info("reading pvalues from stdin")
        all_term2pvalue, all_term2log2fold = readPValues( options.stdin )
    else:
        E.info("reading pvalues from %s " % options.filename_pvalues)
        with IOTools.openFile( options.filename_pvalues ) as infile:
            all_term2pvalue, all_term2log2fold = readPValues( infile )
        
    E.info( "read %i pvalue and %i fold assignments" % ( len(all_term2pvalue), len(all_term2log2fold) ))

    for test_ontology in options.ontology:
        
        E.info( "working on namespace %s" % test_ontology)

        namespace = map_ontology2namespace[test_ontology]

        ################################################################
        # compute or load ancestry information for all nodes
        fn = buildFilename( test_ontology, "ancestors", "tsv.gz" )
        if os.path.exists( fn ):
            E.info( "reading ancestor information from %s" % fn )
            with IOTools.openFile( fn ) as infile:
                ancestors = readAncestors( infile )
        else:
            graph = readGOGraph( options.filename_obo, namespace )
            ancestors = computeAncestors( graph, graph.nodes() )
            with IOTools.openFile( fn, "w" ) as outfile:
                outfile.write("goid\tancestors\n" )
                outfile.write("".join( [ "%s\t%s\n" % (x,";".join(v)) for x,v in ancestors.iteritems()] ) )

        #########################################
        graph = readGOGraph( options.filename_obo, namespace )

        #########################################
        E.info("reading gene to go assignments from %s " % options.filename_go)
        all_gene2gos, all_go2infos = GO.ReadGene2GOFromFile( IOTools.openFile( options.filename_go ), 
                                                             synonyms  = graph.synonyms,
                                                             obsolete = graph.obsoletes )

        #########################################
        # filter pvalues
        synonyms = graph.synonyms 
        term2pvalue = dict([ (synonyms.get(x,x),y) for x,y in all_term2pvalue.iteritems() ])
        term2log2fold = dict([ (synonyms.get(x,x),y) for x,y in all_term2log2fold.iteritems() ])

        if len(term2pvalue) == 0:
            E.warn( "no data - no output produced" )
            E.Stop()
            return
        
        go2info = all_go2infos[test_ontology]
        go2info = collections.defaultdict( str, dict( [(x.mGOId, x.mDescription) for x in go2info.values()] ) )
        gene2gos = all_gene2gos[test_ontology]

        if options.filename_bg:
            background = IOTools.readList( IOTools.openFile( filename_bg )) 
        else:
            background = list(gene2gos.keys())

        # invert mapping, at the same time insert missing ancestral counts
        go2genes = GO.buildGO2Genes( gene2gos, ancestors = ancestors )

        ################################################################
        # compute or load term frequences
        fn = buildFilename( test_ontology, "termfrequencies", "tsv.gz" )
        if os.path.exists( fn ):
            E.info( "reading term frequencies from %s" % fn )
            with IOTools.openFile( fn ) as infile:
                counts, p = readTermFrequencies( infile )
        else:
            counts, p = computeTermFrequencies( graph, go2genes )
            with IOTools.openFile( fn, "w" ) as outfile:
                outfile.write( "goid\tfrequency\tcounts\n" )
                for k in counts.keys():
                    outfile.write("%s\t%f\t%s\n" % (k, p[k], ";".join(counts[k]) ))

        # compute or load simrel matrix
        fn = buildFilename( test_ontology, "simrel", "matrix.gz" )
        if os.path.exists( fn ):
            E.info( "reading simrel matrix from %s" % fn )
            with IOTools.openFile( fn, "r") as infile:
                matrix, terms = readSimrelMatrix( infile )
        else:
            matrix, terms = buildSimrelMatrix( go2genes, go2info, ancestors, p )
            E.info("writing simrel matrix to file" )
            with IOTools.openFile( fn, "w" ) as outfile:
                IOTools.writeMatrix( outfile, matrix, terms, terms )
            
        fn = buildFilename( test_ontology, "cluster", "matrix.gz" )

        if os.path.exists( fn ):
            E.info( "reading clustered matrix matrix from %s" % fn )
            with IOTools.openFile( fn, "r") as infile:
                matrix, terms = readSimrelMatrix( infile )
        else:
            matrix, terms, clusters = clusterSimrelMatrix( matrix, terms, 
                                                           ancestors, p, counts, 
                                                           term2pvalue, go2info,
                                                           options )

            if matrix == None:
                E.warn( "%s: no output" % test_ontology)
                continue
            
            E.info( "after clustering: %i x %i matrix" % (len(terms), len(terms)))
            with IOTools.openFile( fn, "w" ) as outfile:
                IOTools.writeMatrix( outfile, matrix, terms, terms )

            fn = buildFilename( test_ontology, "cluster", "tsv")

            with IOTools.openFile( fn, "w" ) as outfile:
                outfile.write( "\t".join( ("rep", "mem", "pvalue", "frequency", "description") )+ "\n" )
                for rep, members in clusters.iteritems():
                    for mem in members:
                        outfile.write("%s\t%s\t%f\t%f\t%s\n" % (
                                rep, mem, term2pvalue[mem], p[mem], go2info.get(mem, "na")))


        # visualize output
        rows = len(terms)
            
        # revigo
        #     color = pvalue, size = frequency, egdewith = simrel score
        # me:
        #     size  = pvalue
        #     color = fold (blueish for depletion, reddish for enrichment), 
        #     edgewith = can't be done with networkx? (only edgecolor)

        # build graph from matrix
        term_graph = networkx.Graph()
        figure = plt.figure()

        term_graph.add_nodes_from( terms )
        for i,j in itertools.combinations( xrange( rows), 2 ):
            w = weight=matrix[i][j]
            if w < 0.1: continue
            term_graph.add_edge( terms[i], terms[j], weight = w )

        layout = networkx.spring_layout( term_graph )

        if options.reverse_palette:
            color_scheme = eval( "pylab.cm.%s_r" % options.palette)                    
        else:
            color_scheme = eval( "pylab.cm.%s" % options.palette)
        
        labels = dict( [ (x, go2info.get(x,x)) for x in terms] )
        node_size = [ -20.0 * math.log(term2pvalue[x],10) for x in term_graph.nodes() ]
        for x in term2pvalue.keys():
            term2log2fold[x] = numpy.random.normal(0, 5)
        node_color = [ term2log2fold[x] for x in term_graph.nodes() ]

        networkx.draw_networkx_edges( term_graph, layout, edge_color='0.50' )

        vmin = min(node_color)
        vmax = max(node_color)
        vmin = min( vmin, -vmax )
        vmax = max( vmax, -vmin )

        networkx.draw_networkx_nodes( term_graph, layout, 
                                      node_size = node_size,
                                      node_color = node_color,
                                      cmap = color_scheme,
                                      alpha = 0.8,
                                      vmin = vmin,
                                      vmax = vmax,
                                      linewidths = 0.5 )

        positions = dict( [ (x, numpy.array( (y[0],y[1]-0.02)) ) for x,y in layout.iteritems() ] )
        
        networkx.draw_networkx_labels( term_graph, 
                                       positions, 
                                       labels = labels,
                                       font_size = 5,
                                       alpha = 0.5 )

        # turn of tick marks
        gca = figure.gca()
        gca.set_xticklabels( [], visibile = False )
        gca.set_yticklabels( [], visibile = False )
        
        fn = buildFilename( test_ontology, "graph", "svg" )
        plt.colorbar()
        plt.savefig( fn )
        
        fn = buildFilename( test_ontology, "graph", "tsv.gz" )
        with IOTools.openFile( fn, "w") as outfile:
            outfile.write("goid\tx\ty\tpvalue\tl2fold\tdescription\n" )
            for term in terms:
                outfile.write("%s\t%f\t%f\t%f\t%f\t%s\n" % \
                                  (term, 
                                   layout[term][0],
                                   layout[term][1],
                                   term2pvalue[term],
                                   term2log2fold[term],
                                   go2info.get(term,"na") ) )

    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
bam2geneprofile.py - build meta-gene profile for a set of transcripts/genes
===========================================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script takes a :term:`gtf` formatted file and computes meta-gene profiles
over various annotations derived from the :term:`gtf` file. 

A meta-gene profile is an abstract genomic entity over which reads stored in a 
:term:`bam` formatted file have been counted. A meta-gene might be an idealized
eukaryotic gene (upstream, exonic sequence, downstream) or any other genomic landmark 
of interest such as transcription start sites.

The script can be used to visualize binding profiles of a chromatin mark in gene 
bodies, binding of transcription factors in promotors or 3' bias in RNASeq data.

This script is designed with a slight emphasis on RNA-Seq datasets. For example, it
takes care of spliced reads, by using the CIGAR string in the BAM file to accurately
define aligned bases (when the --base-accurate-off is not specified, currently it is not specified by default).

Alternatively, for the purpose of visualizing binding profiles of transcription factors
ChIP-Seq without the need to use any genomic annotations (ENSEMBL, or refseq), you may
also consider using :doc:`bam2peakshape`, which is
designed with a slight emphasis on Chip-Seq datasets, for example, :doc:`bam2peakshape` is able to
center the counting window to the summit of every individual peak. :doc:`bam2peakshape` is also able
to: (1) plot the control RNA-Seq library to enable side-by-side comparison; (2) randomize
the given regions to provide a semi-control.

Background
---------- 

The :file:`bam2geneprofile.py` script reads in a set of transcripts from a :term:`gtf` 
formatted file. For each transcript, overlapping reads from the provided :term:`bam` file 
are collected. The counts within the transcript are then mapped onto the meta-gene
structure and counts are aggregated over all transcripts in the :term:`gtf` file.

:term:`Bam` files need to be sorted by coordinate and indexed. 

A meta-gene structure has two components - regions of variable size, such as exons, introns, etc,
which nevertheless have a fixed start and end coordinate in a transcript. The other component
are regions of fixed width, such a regions of a certain size upstream or downstream of a landmark such as
a transcription start site.

The size of the former class, regions of variable size, can be varied with ``--resolution`` options.
For example, the option ``--resolution-upstream-utr=1000`` will create a meta-gene with 
a 1000bp upstream UTR region. UTRs that are larger will be compressed, but UTRs that are smaller, will
be stretched to fit the 1000bp meta-gene UTR region.

The size of fixed-width regions can be set with ``--extension`` options. For example,
the options ``--extension-upstream`` will set the size of the uptsream extension region
to 1000bp. Note that no scaling is required when counting reads towards the meta-gene profile.

Options
-------

The behaviour of the script can be modified by several options.

Profiles
++++++++

Different profiles are accessible through the ``--counter`` option. Multiple
counters can be applied at the same time. While ``upstream`` and ``downstream``
typically have a fixed size, the other regions such as ``CDS``, ``UTR`` will be
scaled to a common size.

utrprofile
    UPSTREAM - UTR5 - CDS - UTR3 - DOWNSTREAM
    gene models with UTR. Separate the coding section from the non-coding part.

geneprofile 
    UPSTREAM - EXON - DOWNSTREAM
    simple exonic gene models

geneprofilewithintrons 
    UPSTREAM - EXON - INTRON - DOWNSTREAM
    gene models containing also intronic sequence, only correct if used with --base-accuracy
     
geneprofileabsolutedistancefromthreeprimeend
    UPSTREAM - EXON(AbsoluteFromThreePrimeEnd) - INTRON(AbsoluteFromThreePrimeEnd) - DOWNSTREAM
    For EXON region, the script counts over the mRNA transcript only, skipping introns. 
    Designed to visualize the 3 prime bias in RNASeq data, only correct if used with --base-accuracy
    
tssprofile
    UPSTREAM - DOWNSTREAM
    transcription start/stop sites

intervalprofile 
    UPSTREAM - INTERVAL - DOWNSTREAM
    Similar to geneprofile, but count over the complete span of the gene
    (including introns).

midpointprofile
    UPSTREAM  - DOWNSTREAM
    aggregate over midpoint of gene model

Genes versus transcripts
++++++++++++++++++++++++

The default is to collect reads on a per-transcript level. Alternatively, the script can
merge all transcripts of a gene into a single virtual transcript. Note that this virtual transcript
might not be a biologically plausible transcript. It is usually better to provide :file:`bam2geneprofile.py`
with a set of representative transcripts per gene in order to avoid up-weighting genes with
multiple transcripts.

Meta-gene structures
++++++++++++++++++++

The script provides a variety of different meta-gene structures (``--methods``).

Normalization
+++++++++++++

Normalization can be applied in two stages of the computation.

Before adding counts to the meta-gene profile, the profile for the individual
transcript can be normalized. Without normalization, highly expressed genes
will contribute more to the meta-gene profile than lowly expressed genes.
With normalization, each gene contributes an equal amount (``-normalize``).

When outputting the meta-gene profile, the meta-gene profile itself can be normalized.
Normalization a profile can help comparing the shapes of profiles between different
experiments independent of the number of reads or transcripts used in the construction
of the meta-gene profile (``--normalize-profile``).

Bed and wiggle files
++++++++++++++++++++

The densities can be computed from :term:`bed` or :term:`wiggle` formatted files.
If a :term:`bed` formatted file is supplied, it must be compressed with and indexed with :file:`tabix`.
.. todo::
   
   * paired-endedness is ignored. Both ends of a paired-ended read are treated 
     individually.


Usage
-----

The following command will generate the gene profile from RNA-Seq data:

   python bam2geneprofile.py --bamfile=rnaseq.bam -g geneset.gtf.gz 
                     --method=geneprofilewithintrons  --reporter=gene 
                     --resolution-cds=1400 --resolution-introns=2000

The output will contain read coverage over genes, merging all transcripts within
a gene into a single chain of exons. The profile will contain four separate
segments:

1. the upstream region of a gene (default = 1000bp)
2. the transcribed region of a gene. The transcribed region of every gene will be 
   scaled to 1400 bp (``--resolution-cds=1400``), 
   shrinking longer transcripts and expanding shorter transcripts.
3. the intronic regions of a gene. These will be scaled to 2kb (``-resolution-introns=2000``).
4. the downstream region of a gene (default = 1000bp).

Type::

   python bam2geneprofile.py --help

for command line help.

Command line options
--------------------

'''

import os
import sys
import re
import optparse
import collections
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import pysam
import CGAT.GTF as GTF
import numpy

try:
    import pyximport
    pyximport.install(build_in_temp=False)
    import _bam2geneprofile
except ImportError:
    import CGAT._bam2geneprofile as _bam2geneprofile


from bx.bbi.bigwig_file import BigWigFile

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                                    usage = globals()["__doc__"] )

    parser.add_option( "-m", "--method", dest="methods", type = "choice", action = "append",
                       choices = ("geneprofile", "tssprofile", "utrprofile", 
                                  "intervalprofile", "midpointprofile",
                                  "geneprofilewithintrons", 
                                  "geneprofileabsolutedistancefromthreeprimeend",
                                  ),
                       help = 'counters to use. Counters describe the meta-gene structure to use '
                       '[%default]. \n Note using geneprofilewithintrons, or geneprofileabsolutedistancefromthreeprimeend will automatically turn on the --base-accuracy option' )

    parser.add_option( "-b", "--bamfile", "--bedfile", "--bigwigfile", dest="infiles", 
                       metavar = "BAM",
                       type = "string", action = "append",
                       help = "BAM/bed/bigwig files to use. Do not mix different types"
                              "[%default]" )

    parser.add_option( "-g", "--gtffile", dest="gtffile", type = "string",
                       metavar = "GTF",
                       help = "GTF file to use. "
                              "[%default]" )

    parser.add_option( "-n", "--normalization", dest="normalization", type = "choice",
                       choices = ("none", "max", "sum", "total-max", "total-sum"),
                       help = """normalization to apply on each transcript profile before adding to meta-gene profile. 

The options are:

* none: no normalization
* sum: sum of counts within a region
* max: maximum count within a region
* total-sum: sum of counts across all regions
* total-max: maximum count in all regions
[%default]""" )

    parser.add_option( "-p", "--normalize-profile", dest="profile_normalizations", type = "choice", action="append",
                       choices = ("none", "area", "counts"),
                       help = """normalization to apply on meta-gene profile normalization. 

The options are:
* none: no normalization
* area: normalize such that the area under the meta-gene profile is 1.
* counts: normalize by number of features (genes,tss) that have been counted
[%default]""" )

    parser.add_option( "-r", "--reporter", dest="reporter", type = "choice",
                       choices = ("gene", "transcript"  ),
                       help = "report results for genes or transcripts."
                              " When 'genes` is chosen, exons across all transcripts for"
                              " a gene are merged. When 'transcript' is chosen, counts are"
                              " computed for each transcript separately with each transcript"
                              " contributing equally to the meta-gene profile."
                              " [%default]" )

    parser.add_option( "-i", "--shift", dest="shifts", type = "int", action = "append",
                       help = "shift reads in :term:`bam` formatted file before computing densities (ChIP-Seq). "
                       "[%default]" )

    parser.add_option( "-a", "--merge-pairs", dest="merge_pairs", action = "store_true",
                       help = "merge pairs in :term:`bam` formatted file before computing"
                              " densities (ChIP-Seq)."
                              "[%default]" )

    parser.add_option( "-u", "--base-accuracy", dest="base_accuracy", action = "store_true",
                       help = "compute densities with base accuracy. The default is to"
                              " only use the start and end of the aligned region (RNA-Seq)"
                              " [%default]" )

    parser.add_option( "-e", "--extend", dest="extends", type = "int", action = "append",
                       help = "extend reads in :term:`bam` formatted file (ChIP-Seq). "
                              "[%default]" )

    parser.add_option("--resolution-upstream", dest="resolution_upstream", type = "int",
                       help = "resolution of upstream region in bp "
                              "[%default]" )

    parser.add_option("--resolution-downstream", dest="resolution_downstream", type = "int",
                       help = "resolution of downstream region in bp "
                              "[%default]" )

    parser.add_option("--resolution-upstream-utr", dest="resolution_upstream_utr", type = "int",
                       help = "resolution of upstream UTR region in bp "
                              "[%default]" )

    parser.add_option("--resolution-downstream-utr", dest="resolution_downstream_utr", type = "int",
                       help = "resolution of downstream UTR region in bp "
                              "[%default]" )

    parser.add_option("--resolution-cds", dest="resolution_cds", type = "int",
                       help = "resolution of cds region in bp "
                              "[%default]" )

    parser.add_option("--resolution-introns", dest="resolution_introns", type = "int",
                       help = "resolution of introns region in bp "
                              "[%default]" )
                              
    parser.add_option("--resolution-exons-absolute-distance-topolya", dest="resolution_exons_absolute_distance_topolya", type = "int",
                       help = "resolution of exons absolute distance topolya in bp "
                              "[%default]" )                    
                       
    parser.add_option("--resolution-introns-absolute-distance-topolya", dest="resolution_introns_absolute_distance_topolya", type = "int",
                       help = "resolution of introns absolute distance topolya in bp "
                              "[%default]" )
                       
    parser.add_option("--extension-exons-absolute-distance-topolya", dest="extension_exons_absolute_distance_topolya", type = "int",
                       help = "extension for exons from the absolute distance from the topolya in bp"
                              "[%default]" )

    parser.add_option("--extension-introns-absolute-distance-topolya", dest="extension_introns_absolute_distance_topolya", type = "int",
                       help = "extension for introns from the absolute distance from the topolya in bp"
                              "[%default]" )                
                              
    parser.add_option("--extension-upstream", dest = "extension_upstream", type = "int",
                       help = "extension upstream from the first exon in bp"
                              "[%default]" )

    parser.add_option("--extension-downstream", dest = "extension_downstream", type = "int",
                       help = "extension downstream from the last exon in bp"
                              "[%default]" )

    parser.add_option("--extension-inward", dest="extension_inward", type = "int",
                       help = "extension inward from a TSS start site in bp"
                              "[%default]" )

    parser.add_option("--extension-outward", dest="extension_outward", type = "int",
                       help = "extension outward from a TSS start site in bp"
                              "[%default]" )
                       
    parser.add_option("--scale-flank-length", dest="scale_flanks", type = "int",
                       help = "scale flanks to (integer multiples of) gene length"
                              "[%default]" )

    parser.set_defaults(
        remove_rna = False,
        ignore_pairs = False,
        input_reads = 0,
        force_output = False,
        bin_size = 10,
        extends = [],
        shifts = [],
        sort = [],
        reporter = "transcript",
        resolution_cds = 1000,
        resolution_introns = 1000,
        resolution_exons_absolute_distance_topolya = 3000,    #3kb is a good balance of seeing long enough 3 prime bias and not omit too many genes. Tim 31th Aug 2013
        resolution_introns_absolute_distance_topolya = 500,   #introns is only for assess the noise level, thus do ont need a long region, a long region has the side effect of omit more genes. Tim 31th Aug 2013
        extension_exons_absolute_distance_topolya = 3000,     #3kb is a good balance of seeing long enough 3 prime bias and not omit too many genes. Tim 31th Aug 2013
        extension_introns_absolute_distance_topolya = 500,    #introns is only for assess the noise level, thus do ont need a long region, a long region has the side effect of omit more genes. Tim 31th Aug 2013
        resolution_upstream_utr = 1000,
        resolution_downstream_utr = 1000,
        resolution_upstream = 1000,
        resolution_downstream = 1000,
        # mean length of transcripts: about 2.5 kb
        extension_upstream = 2500,
        extension_downstream = 2500,
        extension_inward = 3000,
        extension_outward = 3000,
        plot = True,
        methods = [],
        infiles = [],
        gtffile = None,
        profile_normalizations = [],
        normalization = None,
        scale_flanks = 0,
        merge_pairs = False,
        min_insert_size = 0,
        max_insert_size = 1000,
        base_accuracy = False,
        )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv, add_output_options = True )

    # Keep for backwards compatability
    if len(args) == 2:
        infile, gtf = args
        options.infiles.append(infile)
        options.gtffile = gtf

    if not options.gtffile:
        raise ValueError("no GTF file specified" )

    if len(options.infiles) == 0:
        raise ValueError("no bam/wig/bed files specified" )
    
    for methodsRequiresBaseAccuracy in ["geneprofilewithintrons", 
                                        "geneprofileabsolutedistancefromthreeprimeend",
                                       ]:
        # If you implemented any methods that you do not want the spliced out introns 
        # or exons appear to be covered by non-existent reads, it is better you let those
        # methods imply --base-accurarcy by add them here.
        if methodsRequiresBaseAccuracy in options.methods:
            options.base_accuracy=True

    if options.reporter == "gene":
        gtf_iterator = GTF.flat_gene_iterator( GTF.iterator( IOTools.openFile( options.gtffile ) ) )
    elif options.reporter == "transcript":
        gtf_iterator = GTF.transcript_iterator( GTF.iterator( IOTools.openFile( options.gtffile ) ) )
        
    # Select rangecounter based on file type
    if len(options.infiles) > 0:
        if options.infiles[0].endswith( ".bam" ):
            bamfiles = [ pysam.Samfile( x, "rb" ) for x in options.infiles ]
            format = "bam"
            if options.merge_pairs:
                range_counter = _bam2geneprofile.RangeCounterBAM( bamfiles, 
                                                                  shifts = options.shifts, 
                                                                  extends = options.extends,
                                                                  merge_pairs = options.merge_pairs,
                                                                  min_insert_size = options.min_insert_size,
                                                                  max_insert_size = options.max_insert_size )
            elif options.shifts or options.extends:
                range_counter = _bam2geneprofile.RangeCounterBAM( bamfiles, 
                                                                  shifts = options.shifts, 
                                                                  extends = options.extends )
            elif options.base_accuracy:
                range_counter = _bam2geneprofile.RangeCounterBAMBaseAccuracy( bamfiles )
            else:
                range_counter = _bam2geneprofile.RangeCounterBAM( bamfiles )
            
                                                              
        elif options.infiles[0].endswith( ".bed.gz" ):
            bedfiles = [ pysam.Tabixfile( x ) for x in options.infiles ]
            format = "bed"
            range_counter = _bam2geneprofile.RangeCounterBed( bedfiles )

        elif options.infiles[0].endswith( ".bw" ):
            wigfiles = [ BigWigFile(file=open(x)) for x in options.infiles ]
            format = "bigwig"
            range_counter = _bam2geneprofile.RangeCounterBigWig( wigfiles )

        else:
            raise NotImplementedError( "can't determine file type for %s" % bamfile )

    counters = []
    for method in options.methods:
        if method == "utrprofile":
            counters.append( _bam2geneprofile.UTRCounter( range_counter, 
                                                          options.resolution_upstream,
                                                          options.resolution_upstream_utr,
                                                          options.resolution_cds,
                                                          options.resolution_downstream_utr,
                                                          options.resolution_downstream,
                                                          options.extension_upstream,
                                                          options.extension_downstream ) )
        elif method == "geneprofile":
            counters.append( _bam2geneprofile.GeneCounter( range_counter, 
                                                           options.resolution_upstream,
                                                           options.resolution_cds,
                                                           options.resolution_downstream,
                                                           options.extension_upstream,
                                                           options.extension_downstream,
                                                           options.scale_flanks ) )
        elif method == "geneprofilewithintrons":
            counters.append( _bam2geneprofile.GeneCounterWithIntrons( range_counter, 
                                                           options.resolution_upstream,
                                                           options.resolution_cds,
                                                           options.resolution_introns,
                                                           options.resolution_downstream,
                                                           options.extension_upstream,
                                                           options.extension_downstream,
                                                           options.scale_flanks ) )

        elif method == "geneprofileabsolutedistancefromthreeprimeend":
            counters.append( _bam2geneprofile.GeneCounterAbsoluteDistanceFromThreePrimeEnd( range_counter, 
                                                           options.resolution_upstream,                                                           
                                                           options.resolution_downstream,
                                                           options.resolution_exons_absolute_distance_topolya,
                                                           options.resolution_introns_absolute_distance_topolya,
                                                           # options.resolution_exons_absolute_distance_tostartsite,   
                                                           # options.resolution_introns_absolute_distance_tostartsite,             
                                                           # Tim 31th Aug 2013: a possible feature for future,  if five prime bias is of your interest. 
                                                           #(you need to create another class). It is not very difficult to derive from this class, but is not implemented yet
                                                           # This future feature is slightly different the TSS profile already implemented, because in this future feature introns are skipped, 
                                                           options.extension_upstream,
                                                           options.extension_downstream,
                                                           options.extension_exons_absolute_distance_topolya,
                                                           options.extension_introns_absolute_distance_topolya,
                                                           # options.extension_exons_absolute_distance_tostartsite,   
                                                           # options.extension_introns_absolute_distance_tostartsite,                                           
                                                           # Tim 31th Aug 2013: a possible feature for future,  if five prime bias is of your interest. 
                                                           #(you need to create another class). It is not very difficult to derive from this class, but is not implemented yet
                                                           # This future feature is slightly different the TSS profile already implemented, because in this future feature introns are skipped, 
                                                           options.scale_flanks ) )
             

        elif method == "tssprofile":
            counters.append( _bam2geneprofile.TSSCounter( range_counter, 
                                                           options.extension_outward,
                                                           options.extension_inward ) )

        elif method == "intervalprofile":
            counters.append( _bam2geneprofile.RegionCounter( range_counter, 
                                                             options.resolution_upstream,
                                                             options.resolution_cds,
                                                             options.resolution_downstream,
                                                             options.extension_upstream,
                                                             options.extension_downstream ) )

        elif method == "midpointprofile":
            counters.append( _bam2geneprofile.MidpointCounter( range_counter, 
                                                               options.resolution_upstream,
                                                               options.resolution_downstream,
                                                               options.extension_upstream,
                                                               options.extension_downstream ) )

    # set normalization
    for c in counters:
        c.setNormalization( options.normalization )

    E.info( "starting counting with %i counters" % len(counters) )

    _bam2geneprofile.count( counters, gtf_iterator )

    for method, counter in zip(options.methods, counters):
        if not options.profile_normalizations:
            options.profile_normalizations.append( "none" )

        for norm in options.profile_normalizations:
            outfile = IOTools.openFile( E.getOutputFile( counter.name ) + ".%s.tsv.gz" % norm, "w")
            counter.writeMatrix( outfile, normalize=norm )
            outfile.close()
            
        outfile = IOTools.openFile( E.getOutputFile( counter.name ) + ".lengths.tsv.gz", "w")
        counter.writeLengthStats( outfile )
        outfile.close()

    if options.plot:

        import matplotlib
        # avoid Tk or any X
        matplotlib.use( "Agg" )
        import matplotlib.pyplot as plt
        
        for method, counter in zip(options.methods, counters):

            if method in ("geneprofile", "geneprofilewithintrons", "geneprofileabsolutedistancefromthreeprimeend", "utrprofile", "intervalprofile" ):

                plt.figure()
                plt.subplots_adjust( wspace = 0.05)
                max_scale = max( [max(x) for x in counter.aggregate_counts ] )

                for x, counts in enumerate( counter.aggregate_counts ):
                    plt.subplot( 5, 1, x+1)
                    plt.plot( range(len(counts)), counts )
                    plt.title( counter.fields[x] )
                    plt.ylim( 0, max_scale )

                figname = counter.name + ".full"
                
                fn = E.getOutputFile( figname ) + ".png"
                plt.savefig( os.path.expanduser(fn) )

                plt.figure()

                points = []
                cuts = []
                for x, counts in enumerate( counter.aggregate_counts ):
                    points.extend( counts )
                    cuts.append( len( counts ) )
                                 
                plt.plot( range(len(points)), points )
                xx,xxx = 0, []
                for x in cuts:
                    xxx.append( xx + x // 2 )
                    xx += x
                    plt.axvline( xx, color = "r", ls = "--" )

                plt.xticks( xxx, counter.fields )

                figname = counter.name + ".detail"
                
                fn = E.getOutputFile( figname ) + ".png"
                plt.savefig( os.path.expanduser(fn) )

            elif method == "tssprofile":

                plt.figure()
                plt.subplot( 1, 3, 1)
                plt.plot( range(-options.extension_outward, options.extension_inward), counter.aggregate_counts[0] )
                plt.title( counter.fields[0] )
                plt.subplot( 1, 3, 2)
                plt.plot( range(-options.extension_inward, options.extension_outward), counter.aggregate_counts[1] )
                plt.title( counter.fields[1] )
                plt.subplot( 1, 3, 3)
                plt.title( "combined" )
                plt.plot( range(-options.extension_outward, options.extension_inward), counter.aggregate_counts[0] )
                plt.plot( range(-options.extension_inward, options.extension_outward), counter.aggregate_counts[1] )
                plt.legend( counter.fields[:2] )

                fn = E.getOutputFile( counter.name ) + ".png"
                plt.savefig( os.path.expanduser(fn) )

            elif method == "midpointprofile":

                plt.figure()
                plt.plot( numpy.arange(-options.resolution_upstream, 0), counter.aggregate_counts[0] )
                plt.plot( numpy.arange(0, options.resolution_downstream), counter.aggregate_counts[1] )

                fn = E.getOutputFile( counter.name ) + ".png"
                plt.savefig( os.path.expanduser(fn) )
        
    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )

    



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
graph_map_links.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python graph_map_links.py --help

Type::

   python graph_map_links.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import getopt
import tempfile
import time
import popen2
import optparse
import hashlib

USAGE="""python %s [OPTIONS] < graph.in > graph.out

Version: $Id: graph_map_links.py 2782 2009-09-10 11:40:29Z andreas $

Map blast links.

Options:
-h, --help                      print this message.
-v, --verbose=                  loglevel.
-q, --map-query=                map to use for query
-s, --map-sbjct=                map to use for sbjct
-m, --multiple                  write multiple matches
-k, --keep-unmapped             keep entries not mapped
-i, --identity-map              mapping is done by identities
""" % sys.argv[0]

import CGAT.Experiment as E
import CGAT.BlastAlignments as BlastAlignments

def ReadIdentityMap( infile):
    """read identity map.

    multiple entries can either be separated over several lines
    or concatenated by semicolon in the same line.
    """
    
    m = {}
    for line in infile:

        if line[0] == "#": continue
        olds, news = map(lambda x: x.split(";"), line[:-1].split("\t")[:2])

        for old in olds:
            if old not in m: m[old] = []
            for new in news:
                m[old].append( new )
    return m

##-------------------------------------------------------------------------------
if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: graph_map_links.py 2782 2009-09-10 11:40:29Z andreas $")

    parser.add_option("-q", "--map-query", dest="filename_map_query", type="string",
                      help="filename with queries to map." )
    parser.add_option("-s", "--map-sbjct", dest="filename_map_sbjct", type="string",
                      help="filename with queries to map." )
    parser.add_option("-m", "--multiple", dest="multiple", action="store_true",
                      help="map multiple options." )
    parser.add_option("-k", "--keep-unmapped", dest="keep_unmapped", action="store_true",
                      help="keep unmapped entries." )
    parser.add_option("-i", "--map-identity", dest="map_identity", action="store_true",
                      help="use identity maps (only identifier change)." )
    parser.add_option("-n", "--non-redundant", dest="non_redundant", action="store_true",
                      help="write only unique links (requires a lot of memory for large graphs).")
    parser.set_defaults( \
        filename_map_query = None,
        filename_map_sbjct = None,
        multiple = False,
        keep_unmapped = False,
        map_identity = False,
        report_step = 1000000,
        non_redundant = False)
        
    (options, args) = E.Start( parser )

    if options.filename_map_query:
        infile = open(options.filename_map_query, "r")        
        if options.map_identity:
            map_query = ReadIdentityMap( infile )
        else:
            map_query = BlastAlignments.ReadMap( infile, options.multiple )
        infile.close()
        if options.loglevel >= 1:
            print "# read maps for %i queries." % len(map_query)
    else:
        map_query = None

    if options.filename_map_sbjct:
        if options.filename_map_sbjct == options.filename_map_query:
            map_sbjct = map_query
        else:
            infile = open(options.filename_map_sbjct, "r")                
            if options.map_identity:
                map_sbjct = ReadIdentityMap( infile )            
            else:
                map_sbjct = BlastAlignments.ReadMap( open(options.filename_map_sbjct, "r"), options.multiple)
            infile.close()
        if options.loglevel >= 1:
            print "# read maps for %i sbjcts." % len(map_sbjct)
    else:
        map_sbjct = None

    nfailed = 0
    ninput = 0
    nskipped = 0
    noutput = 0

    # number of identical/mapped links
    nsame, nmapped = 0, 0

    printed = {}
    
    map = BlastAlignments.Map()

    for line in sys.stdin:
        
        if line[0] == "#": continue

        data = line[:-1].split("\t")

        map.Read( line )
        skip = False
        ninput += 1

        if options.loglevel >= 2:
            print "#", str(map)

        if options.loglevel >= 2 and ninput % options.report_step == 0:
            sys.stderr.write( "# progress: ninput=%i, noutput=%i, nhash=%i\n" % (ninput, noutput, len(printed)) )
        
        if options.multiple:
            skip = False
            if map_query != None:
                if map.mQueryToken in map_query:
                    mq = map_query[map.mQueryToken]
                else:
                    skip = True
            else:
                mq = [ None ]                                

            if map_sbjct != None:                
                if map.mSbjctToken in map_sbjct:
                    ms = map_sbjct[map.mSbjctToken]
                else:
                    skip = True
            else:
                ms = [ None ]
                
            if skip:
                nskipped += 1
                continue
            
            if options.map_identity:

                ## only if non_redundant is set, do global comparison
                if not options.non_redundant: printed = {}
                            
                new_map = map.GetClone()
                do_redundant = len(mq) > 1 or len(ms) > 1                
                for q in mq:
                    for s in ms:
  
                        new_map.mQueryToken = q
                        new_map.mSbjctToken = s

                        ## check for non-redundant links for 1:many or many:many mappings
                        if do_redundant:
                            key = "%s-%i-%i-%s-%i-%i" % (new_map.mQueryToken, new_map.mQueryFrom, new_map.mQueryTo,
                                                         new_map.mSbjctToken, new_map.mSbjctFrom, new_map.mSbjctTo)
                        
                            # hash key to save space
                            hkey = hashlib.md5(key).digest()

                            if hkey in printed: continue
                            
                            printed[hkey] = 1

                        print string.join( [str(new_map)]+ data[9:], "\t")
                        noutput += 1
                        if new_map.mQueryToken == map.mQueryToken and \
                           new_map.mSbjctToken == map.mSbjctToken:
                            nsame += 1
                        else:
                            nmapped += 1
                        
            else:
                for q in mq:
                    for s in ms:
                        new_map = map.GetClone()

                        if options.loglevel >= 2:
                            print "#", str(q)
                            print "#", str(s)

                        is_ok = new_map.MapAlignment( q, s )

                        if not is_ok:
                            nfailed += 1
                        else:
                            print string.join( [str(new_map)]+ data[9:], "\t")
                            noutput += 1
        else:

            if map_query != None:
                if map.mQueryToken in map_query:
                    mq = map_query[map.mQueryToken ]
                else:
                    mq = None
                    skip = True
            else:
                mq = None

            if map_sbjct != None:
                if map.mSbjctToken in map_sbjct:
                    ms = map_sbjct[map.mSbjctToken ]
                else:
                    ms = None
                    skip = True
            else:
                ms = None
                
            if skip and not options.keep_unmapped:
                nskipped += 1
                continue

            if options.loglevel >= 2:
                print "#", str(mq)
                print "#", str(ms)

            if mq or ms:
                is_ok = map.MapAlignment( mq, ms )
            else:
                is_ok = True

            if not is_ok:
                nfailed += 1
            else:
                print string.join( [str(map)]+ data[9:], "\t")
                noutput += 1

    print "# ninput=%i, noutput=%i, nskipped=%i, nfailed=%i, nsame=%i, nmapped=%i" % (\
        ninput, noutput, nskipped, nfailed, nsame, nmapped )
    
    E.Stop()





################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
WrapperDBA.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Code
----

'''
import os, sys, string, re, tempfile, subprocess

"""Wrapper for DBA.
"""

import Genomics

class DBA:

    mOptions = ""
    mExecutable = "dba -quiet -nomatchn -align "
    mEnvironment = "WISECONFIGDIR=/net/cpp-group/src/wise2.2.0/wisecfg; export WISECONFIGDIR;"
    
    def __init__( self, options = ""):
        self.mOptions = options

    def Align( self, s1, s2, result ):

        result.clear()
        
        handle_tmpfile1, filename_tmpfile1 = tempfile.mkstemp()
        handle_tmpfile2, filename_tmpfile2 = tempfile.mkstemp()
        os.write( handle_tmpfile1, ">s1\n%s\n" % (s1))
        os.close( handle_tmpfile1 )
        os.write( handle_tmpfile2, ">s2\n%s\n" % (s2))
        os.close( handle_tmpfile2 )

        statement = string.join( ( "(",
                                   self.mEnvironment, 
                                   self.mExecutable,
                                   self.mOptions,
                                   filename_tmpfile1,
                                   filename_tmpfile2,
                                   ")" ),
                                 " ")

        
        p = subprocess.Popen( statement , 
                              shell=True, 
                              stdin=subprocess.PIPE, 
                              stdout=subprocess.PIPE, 
                              stderr=subprocess.PIPE, 
                              close_fds=True)

        (file_stdout, file_stdin, file_stderr) = (p.stdin, p.stdout, p.stderr)

        file_stdin.close()
        lines = file_stdout.readlines()
        lines_stderr = file_stderr.readlines()
        exit_code = file_stdout.close()
        file_stderr.close()

        if exit_code:
            raise "Error while executing statement %s" % statement

        for line in lines:
            data = re.split("\s+", line[:-1] )
            if len(data) != 6: continue
            x1 = int(data[1]) 
            x2 = int(data[4])
            result.addPair( x1, x2, 0 )

        os.remove( filename_tmpfile1 )
        os.remove( filename_tmpfile2 )
        
        return result
                          
            
        
                                 
        
        
        
        
    
        



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
mali2malis.py - split a multiple alignment into smaller multiple alignments
===========================================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

split a multiple alignment into its constituent parts.

This is the inverse operation performed by :doc:`malis2mali`.

Usage
-----

Example::

   python mali2malis.py --help

Type::

   python mali2malis.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import optparse
import math
import time
import random

import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.Mali as Mali

##------------------------------------------------------------
if __name__ == '__main__':

    parser = E.OptionParser( version = "%prog version: $Id: mali2malis.py 2782 2009-09-10 11:40:29Z andreas $", usage = globals()["__doc__"])

    parser.add_option("-a", "--pattern-mali", dest="pattern_mali", type="string",
                      help="filename pattern for multiple alignment files."  )

    parser.add_option( "--filename-coordinates", dest="filename_coordinates", type="string",
                      help="filename of coordinates that constitute the multiple alignment."  )

    parser.add_option("-i", "--input-format", dest="input_format", type="choice",
                      choices=("plain", "fasta", "clustal" ),
                      help="input format of multiple alignment"  )
    
    parser.add_option("-o", "--output-format", dest="output_format", type="choice",
                      choices=( "fasta", "codeml", "phylip" ),
                      help="output format of multiple alignment"  )

    parser.set_defaults(
        input_format="fasta",
        output_format="fasta",
        filename_coordinates = None,
        pattern_mali= "%s.fasta",
        )

    (options, args) = E.Start( parser )

    ## read coordinates
    if options.filename_coordinates:
        coordinates = []
        for line in open(options.filename_coordinates, "r" ):
            if line[0] == "#": continue
            id, length, position = line[:-1].split("\t")
            if id == "component": continue
            coordinates.append( (id, int(length), int(position) ) )
            
    mali = Mali.Mali()
    mali.readFromFile( sys.stdin, format=options.input_format )

    ids = mali.getIdentifiers()

    ninput, noutput = 0, 0
    
    for id, length, position in coordinates:
        
        ninput += 1
        part_mali = Mali.Mali()

        for x in ids:
            part_mali.addSequence( x, 0, length, mali[x][position:position+length] )

        outfile_name = options.pattern_mali % id

        outfile = open(outfile_name, "w" )

        part_mali.writeToFile( outfile, format = options.output_format )

        noutput += 1

    if options.loglevel >= 1:
        options.stdlog.write("# input=%i, output=%i\n" % (ninput, noutput) )

    E.Stop()
    



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
nmf.py - compute NMF factorization
==================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

Needs to adapted as a libary.

Usage
-----

Example::

   python nmf.py --help

Type::

   python nmf.py --help

for command line help.

Documentation
-------------

Code
----

'''
from numpy import *
from numpy.linalg import norm
from time import time
from sys import stdout

def nmf(V,Winit,Hinit,tol,timelimit,maxiter,log = None):
   """
   (c,W,H) = nmf(V,Winit,Hinit,tol,timelimit,maxiter)
   W,H: output solution
   c: output flag to indicate convergence
   Winit,Hinit: initial solution
   tol: tolerance for a relative stopping condition
   timelimit, maxiter: limit of time and iterations
   log: logfile
   """

   W = Winit
   H = Hinit
   initt = time()

   gradW = dot(W, dot(H, H.T)) - dot(V, H.T)
   gradH = dot(dot(W.T, W), H) - dot(W.T, V)
   initgrad = norm(r_[gradW, gradH.T])

   if log: log.write('# init gradient norm %f\n# progress: ' % initgrad )

   tolW = max(0.001,tol)*initgrad
   tolH = tolW

   converged = False
   for iter in xrange(1,maxiter):
      # stopping condition
      projnorm = norm(r_[gradW[logical_or(gradW<0, W>0)],
                         gradH[logical_or(gradH<0, H>0)]])

      if projnorm < tol*initgrad: 
         if log: log.write( "\n# converged\n" )
         converged = True
         break

      if time() - initt > timelimit: 
         if log: log.write( "\n# time limit reached\n" )
         converged = False
         break

      (W, gradW, iterW) = nlssubprob(V.T,H.T,W.T,tolW,1000, log)
      W = W.T
      gradW = gradW.T

      if iterW==1: tolW = 0.1 * tolW

      (H,gradH,iterH) = nlssubprob(V,W,H,tolH,1000, log)
      if iterH==1: tolH = 0.1 * tolH

      if log and iter % 10 == 0: 
         log.write('.')
         log.flush()
   else:
      converged = False
      if log: log.write( "\n# iteration limit reached\n" )

   if log: log.write( '# Iter = %d Final proj-grad norm %f\n' % (iter, projnorm) )
   return (converged,W,H)

def nlssubprob(V,W,Hinit,tol,maxiter, log = None):
   """
   H, grad: output solution and gradient
   iter: #iterations used
   V, W: constant matrices
   Hinit: initial solution
   tol: stopping tolerance
   maxiter: limit of iterations
   """
 
   H = Hinit
   WtV = dot(W.T, V)
   WtW = dot(W.T, W) 

   alpha = 1; beta = 0.1;
   for iter in xrange(1, maxiter):  
      grad = dot(WtW, H) - WtV
      projgrad = norm(grad[logical_or(grad < 0, H >0)])
      if projgrad < tol: break

      # search step size 
      for inner_iter in xrange(1,20):
         Hn = H - alpha*grad
         Hn = where(Hn > 0, Hn, 0)
         d = Hn-H
         gradd = sum(grad * d)
         dQd = sum(dot(WtW,d) * d)
         suff_decr = 0.99*gradd + 0.5*dQd < 0
         if inner_iter == 1:
            decr_alpha = not suff_decr
            Hp = H
         if decr_alpha: 
            if suff_decr:
               H = Hn
               break
            else:
               alpha = alpha * beta;
         else:
            if not suff_decr or (Hp == Hn).all():
               H = Hp
               break
            else:
               alpha = alpha/beta
               Hp = Hn

      if iter == maxiter:
         if log: log.write( '# max iteration in nlssubprob\n' )
   return (H, grad, iter)



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: maf2psl.py 2879 2010-04-06 14:44:34Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
map2psql.py - convert maf formatted file to a psl formatted file
================================================================

:Author: Andreas Heger
:Release: $Id: maf2psl.py 2879 2010-04-06 14:44:34Z andreas $
:Date: |today|
:Tags: Python

Purpose
-------

convert a maf file to a psl file.

Usage
-----

Type::

   python <script_name>.py --help

for command line help.

Code
----

""" 

import os
import sys
import re
import optparse

import CGAT.Experiment as E
from bx.align import maf
from bx.align.tools import get_components_for_species

import CGAT.Blat as Blat

def threaditer( reader, species ):
    '''iterate over reader and return components for species.'''
    for m in reader:
        components = get_components_for_species( m, species ) 
        if components != None:
            yield components
            
def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: maf2psl.py 2879 2010-04-06 14:44:34Z andreas $", 
                                    usage = globals()["__doc__"] )

    parser.add_option("-q", "--query", dest="query", type="string",
                      help="sequence to use for query [default=%default]."  )

    parser.add_option("-t", "--target", dest="target", type="string",
                      help="sequence to use for target [default=%default]."  )

    parser.set_defaults(
        query = None,
        target = None,
        )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    if options.query == None or options.target == None:
        if len(args) != 2:
            raise ValueError( "please supply two sequence identifiers for query and target" )
        options.query, options.target = args
        
    ## do sth
    ninput, nskipped, noutput = 0, 0, 0

    reader = maf.Reader(options.stdin)
    
    psl = Blat.Match()
    for cc in threaditer( reader, (options.query, options.target) ):
        
        ninput += 1
        query, target = cc

        # treat identfiers like Hsap.GL000223.1
        try:
            data = query.src.split(".")
            qs, qcontig = data[0], ".".join(data[1:])
        except ValueError, msg:
            raise ValueError( "error: could not parse query %s: msg=%s" % (query.src, msg) )
                              
        try:        
            data = target.src.split(".")
            ts, tcontig = data[0], ".".join(data[1:])
        except ValueError, msg:
            raise ValueError( "error: could not parse target %s: msg=%s" % (target.src, msg) )

        assert qs == options.query
        assert ts == options.target
        psl.mQueryId = qcontig
        psl.mSbjctId = tcontig

        psl.fromPair( query.start, query.src_size, query.strand, query.text.upper(),
                       target.start, target.src_size, target.strand, target.text.upper() )
        
        E.debug( "%s\t%s\t%i\t%i\t%s\t%s" % \
                     (qs, qcontig, query.start, query.src_size, query.strand, query.text ) )
        E.debug( "%s\t%s\t%i\t%i\t%s\t%s" % \
                     (ts, tcontig, target.start, target.src_size, target.strand, target.text ) )
        options.stdout.write( "%s\n" % str(psl) )
        noutput += 1

    E.info( "ninput=%i, noutput=%i, nskipped=%i" % (ninput, noutput,nskipped) )

    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )




################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
Mali.py - Tools for multiple alignments 
=======================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Code
----

'''

import re, sys, string, time, math, types, copy, random

class AlignedString:

    mGapChars = ("-", ".")
    mGapChar = "-"
    
    def __init__(self, identifier, fr, to, s, is_mangled = False):
        self.mId = identifier
        self.mFrom = fr
        self.mTo = to
        self.mString = s
        self.mIsMangled = is_mangled

    def __len__(self):
        return len(self.mString)

    def maskColumn(self, column, mask_char = "x"):
        self.maskColumns( self, [column,], mask_char = mask_char)
        
    def maskColumns(self, columns, mask_char = "x"):
        s = list( self.mString )
        for c in columns:
            if s[c] not in self.mGapChars:
                s[c] = mask_char
        self.mString = "".join(s)

    def mapColumns(self, columns, map_function):
        s = list( self.mString )
        for c in columns:
            if s[c] not in self.mGapChars:
                s[c] = map_function(s[c])
        self.mString = "".join(s)

    def takeColumns(self, columns ):
        """take columns.

        not implemented yet: updating of residue numbers.
        """
        s = []
        
        for c in columns:
            s.append(self.mString[c])
            
        self.mString = "".join(s)

    def truncate( self, first, last):
        """truncate aligned string.

        not implemented yet: updating of residue numbers.
        """
        self.mString = self.mString[first:last]

    def insertColumns( self, position, num_columns, char = "-"):
        """insert num_columns columns at position."""
        self.mString = self.mString[:position] + char * num_columns + self.mString[position:]

    def getIdentifier( self ) :
        '''return demangled identifier.'''
        if self.mIsMangled:
            return self.mId.rsplit("_", 1)[0]
        else:
            return self.mId
        

    def getResidueNumber( self, position ):
        x = self.mFrom
        for c in self.mString[:position]:
            if c not in self.mGapChars:
                x += 1

        return x

    def getSequence( self ):
        """return sequence without gaps."""
        return re.sub("[%s]" % "".join(self.mGapChars), "", self.mString)
    
    def getSegments( self, transitions, gap_chars = "-." ):
        """return segments of alignment according to
        transitions.

        Transitions is a list of residue numbers.
        """
        segments = []
        pos = self.mFrom

        first_x = 0
        for x in range(len(self.mString)):

            char = self.mString[x]
            
            if char not in gap_chars:
                if pos in transitions:
                    if first_x != x:
                        segments.append( (first_x, x) )
                    first_x = x
                pos += 1

        ## add last segment, unless it has already been added
        ## because it was part of the transitions
        if first_x != x:
            segments.append( (first_x, len(self.mString) ) )
        elif pos in transitions:
            segments.append( (first_x, len(self.mString) ) )

        return segments

    def threadSequence( self, new_sequence, map_old2new ):
        """thread a new sequence onto this sequence replacing
        characters in this sequence with those in new sequence.

        ``map_old2new`` lists the positions that should be replaced.
        It maps non-gap positions in old to those in new. Non-gap
        positions than are not mapped are set to :attr:`gapchar`.
        """

        s = list(self.mString)
        c = 0
        for x in range(len(s)):
            m = map_old2new.mapRowToCol(c)
            if m >= 0: s[c] = new_sequence[m]
            else: s[c] = self.mGapChar
            if s[c] not in self.mGapChars:
                c += 1
                
        self.mString = "".join(s)
        
class MaliData:

    def __init__(self, line, gap_chars = "-.", mask_chars = "Nn" ):

        self.mNGaps   = len(re.sub("[^%s]" % gap_chars, "", line ))
        self.mNMasked = len(re.sub("[^%s]" % mask_chars, "", line ))
        self.mNAll    = len(line)
        self.mNChars  = self.mNAll - self.mNMasked - self.mNGaps

    def __str__(self ):
        return "%i\t%i\t%i\t%i" % (self.mNAll, self.mNChars, self.mNGaps, self.mNMasked)

class Mali:

    mGapChars = ("-", ".")
    mGapPattern = re.compile("[.-]")
    # character used for gaps
    mGapChar = "-"
    # admissable gap characters
    mGapChars = ("-", ".")
    mMaskChar = "X"
    
    def __init__(self):

        self.mIdentifiers = []
        self.mMali = {}
        self.mLength = 0
        self.mAnnotations = {}
        self.mName = None

        ## set to false, if ranges shall not be output.
        self.mWriteRanges = True

    def __contains__(self, key):
        return key in self.mMali
        
    def __getitem__(self, key):
        return self.mMali[key].mString

    def __len__(self):
        return len(self.mIdentifiers)

    def __delitem__(self, key ):
        del self.mMali[key]
        self.mIdentifiers.remove(key)

    def getClone( self ):
        return copy.deepcopy( self )

    def iteritems( self ):
        return self.mMali.iteritems()

    def items(self):
        return self.mMali.items()
    
    def values(self):
        return self.mMali.values()

    def keys(self):
        return self.mMali.keys()

    def getName(self ):
        return self.mName

    def setName(self, name ):
        self.mName = name

    def getIdentifiers(self):
        return self.mIdentifiers
    
    def getLength(self):
        """deprecated."""
        return self.getNumSequences()

    def getNumSequences( self ):
        return self.__len__()

    def getNumColumns( self ):
        if self.mIdentifiers:
            return len(self.mMali[self.mIdentifiers[0]])
        else:
            return 0
    
    def getWidth(self):
        """deprecated."""
        return self.getNumColumns()

    def rename( self, old_name, new_name ):
        """rename an entry."""
        if old_name not in self.mMali:
            raise KeyError, "%s not in mali" % old_name
        self.mIdentifiers[self.mIdentifiers.index( old_name) ] = new_name
        self.mMali[new_name] = self.mMali[old_name]
        del self.mMali[old_name]

    def isEmpty(self):
        return len(self.mMali) == 0

    def getSequence( self, key ):
        return self.mMali[key]

    def getResidueNumber( self, key, position ):
        """return residue number in sequence key at position position."""
        return self.mMali[key].getResidueNumber( position )

    def setSequence( self, key, sequence ):
        ## TODO: updating of residue numbers
        self.mMali[key].mString = sequence

    def getEntry( self, key ):
        return self.mMali[key]

    def addEntry( self, s ):
        """add an aligned string object."""
        if s.mId in self.mMali.keys():
            raise KeyError( "id %s already in mali" % s.mId )

        self.mIdentifiers.append( s.mId )
        self.mMali[s.mId] = s

    def addSequence( self, id, fr, to, sequence ):

        if to < 0: to = self.countCharacters( sequence)
        s = AlignedString( id, fr, to , sequence )
        self.addEntry( s )

    def countCharacters(self, row ):
        return len(row) - len(self.mGapPattern.findall(row))

    def deleteEntry( self, identifier ):
        if identifier not in self.mMali:
            raise KeyError, "identifier %s not in mali." % identifier
        del self.mMali[identifier]
        self.mIdentifiers.remove(identifier)

    def getColumns( self ):
        """return mali in column orientation."""
        args = map( lambda x: self.mMali[x].mString, self.mIdentifiers)
        return map( lambda x: "".join(x), zip( *args ))

    def getConsensus( self, mark_with_gaps = False ):
        """return consensus string.

        The consensus string returns the most frequent character per column
        that is not a gap.
        If mark_with_gaps is set to True, positions with any gap characater are
        set to gaps.
        """
        columns = self.getColumns()
        seq = []
        for x in range(len(columns)):
            s = columns[x]
            counts = [(a, s.count(a)) for a in set(list(s)).difference( set(self.mGapChar) ) ]
            if mark_with_gaps and self.mGapChar in columns[x]:
                seq.append( self.mGapChar )
            else:
                counts.sort( key=lambda x: -x[1] )
                seq.append( counts[0][0] )
        return "".join(seq)
    
    def readFromFile( self, infile, format = "fasta" ):
        """read multiple alignment from file in various format."""
        
        self.mMali = {}
        self.mIdentifiers = []

        pattern_parse_ranges=re.compile("(\S+)/(\d+)-(\d+)")

        # read profiles - a profile possibly consists of several entries per file
        # so treat it differently
        if format.lower() == "profile":

            while 1:
                line = infile.readline()
                if not line: return False
                if line[0] != "#": break

            if line[0] != ">": raise "expected '>' at as first character in line %s" % line

            try:
                self.mName, length, width = re.match(">profile=(\S+) length=(\d+) width=(\d+)", line ).groups()
            except AttributeError:
                raise "could not parse header line %s" % line
            width = int(width)
            for x in range( 0, width ):
                id = "seq%i" % x
                self.mIdentifiers.append( id )
                line = infile.readline()
                if not line: raise "expected %i sequences, only got %i" % (width, x )
                self.mMali[id] = AlignedString(id, 0, self.countCharacters( line[:-1] ), line[:-1])
            return True
                
        if type(infile) == types.ListType or type(infile) == types.TupleType:
            lines = infile
        else:
            lines = infile.readlines()
        
        if format not in ("stockholm"):
            # save comments 
            self.mComments = filter( lambda x: x[0] == "#", lines)
            lines = filter( lambda x: x[0] != "#", lines)
        else:
            self.mComments = []
            
        ## remove empty lines
        lines = filter( lambda x: x.strip(), lines )
        if not lines:
            raise AttributeError, "empty alignment"

        def getId( id, s ):
            x = pattern_parse_ranges.match( id )
            if x:
                id, fr, to = x.groups()
                fr, to = int(fr) - 1, int(to)
            else:
                fr, to = 0, self.countCharacters( s )
                self.mWriteRanges = False

            return id, fr, to

        ###############################################################################
        if format.lower() == "plain":

            for line in lines:
                if not line.strip(): continue
                data = line[:-1].split("\t")
                id = data[3]
                xid = id
                x = 0
                while xid in self.mMali:
                    xid = id + "-" + str(x)
                    x += 1
                    
                self.addEntry( AlignedString(xid, int(data[0]) - 1, int(data[2]), data[1] ) )

        ###############################################################################
        elif format.lower() == "fasta":
            pattern_identifier = "\S+"
            id = None
            fragments = []
            for line in lines:
                if line[0] == ">":
                    if id:
                        s = re.sub( "\s", "", string.join( fragments, ""))
                        id, fr, to = getId( id, s )
                        self.addEntry( AlignedString(id, fr, to, s) )

                    id = re.search( "^(%s)" % pattern_identifier, line[1:-1]).group(0)
                    fragments = []
                    continue
                fragments.append( line[:-1] )

            s = re.sub( "\s", "", string.join( fragments, ""))                
            id, fr, to = getId( id, s )
            self.addEntry( AlignedString(id, fr, to, s) )

        ###############################################################################
        elif format.lower() == "phylip":
            nsequences, nchars = re.split("\s+", lines[0][:-1].strip())
            nsequences = int(nsequences)
            for line in lines[1:]:
                l = line[:-1].strip()
                if not l: continue
                id, sequence = re.match("(\S+)\s+(.*)", l).groups()
                sequence = re.sub( "\s", "", sequence )                
                if id not in self.mMali:
                    self.mIdentifiers.append(id)
                    self.mMali[id] = []

                self.mMali[id].append( sequence )

            for id, frags in self.mMali.items():
                s = "".join(frags)
                fr, to = 0, self.countCharacters( s )                
                self.mMali[id] = AlignedString( id, fr, to, s )

        ###############################################################################            
        elif format.lower() == "clustal":
            ## skip header line
            del lines[0]
            fragments = {}

            ## prune lines
            lines = map( lambda x: x.strip(), lines )
            ## remove empty lines
            lines = filter( lambda x: len(x[:-1]) > 0, lines )
            
            for line in lines:
                ## remove consensus lines
                if line[0] in ("*", ":"): continue
                
                data = re.split( "\s+", line )
                if len(data) != 2:
                    raise ValueError, "parsing error in line %s" % line
                
                id, fragment = data
                if id not in fragments:
                    fragments[id] = []
                    self.mIdentifiers.append(id)
                    
                fragments[id].append( fragment )

            for id, f in fragments.items():
                s = re.sub( "\s", "", string.join( f, ""))                        
                self.mMali[id] = AlignedString(id, 0, self.countCharacters( s ), s)

        elif format.lower() == "stockholm":
            ## skip header line
            assert lines[0].startswith( "# STOCKHOLM" ), "file is not in stockholm format"
            del lines[0]
            fragments = {}
            annotations = {}
            ## prune lines
            lines = map( lambda x: x.strip(), lines )
            ## remove empty lines
            lines = filter( lambda x: len(x[:-1]) > 0, lines )
            
            for line in lines:
                data = re.split( "\s+", line )
                
                if data[0] == "//": break
                
                if line[0] == '#':
                    if data[0] == "#=GC":
                        id, fragment = data[1:3]
                    else:
                        self.mComments.append( line )
                        continue
                    if id not in annotations:
                        annotations[id] = []
                    annotations[id].append( fragment )
                else:
                    
                    if len(data) > 2:
                        raise ValueError, "parsing error in line %s" % line
                    elif len(data) == 1:
                        ## treat empty alignments/lines
                        id = data[0]
                        fragment = ""
                    else:
                        id, fragment = data
                    
                    if id not in fragments:
                        fragments[id] = []
                        self.mIdentifiers.append(id)
                    
                    fragments[id].append( fragment )

            n = []
            for id in self.mIdentifiers:
                f = fragments[id]

                s = re.sub( "\s", "", string.join( f, ""))
                x = pattern_parse_ranges.match( id )
                if x:
                    id, fr, to = x.groups()
                    fr, to = int(fr) - 1, int(to)
                else:
                    fr, to = 0, self.countCharacters( s )

                n.append(id)                    
                self.mMali[id] = AlignedString(id, fr, to, s)
            self.mIdentifiers = n
            
            for id, f in annotations.items():
                s = re.sub( "\s", "", string.join( f, ""))
                annotations[id] = s
            self.mAnnotations = annotations
        else:
            raise "unknown alignment format %s" % format

        if len(self.mMali) == 0:
            self.mLength = 0
        else:
            self.mLength = min( map( lambda x: len(x.mString), self.mMali.values() ) )
        
    def writeToFile( self, outfile, write_ranges = True, format = "plain", options = None):
        """write alignment to file.

        If options is given, these lines are output into the multiple alignment.
        """
        if format == "plain-fasta":
            format = "fasta"
            write_ranges = False

        write_ranges = write_ranges and self.mWriteRanges

        if format == "plain":
            for identifier in self.mIdentifiers:
                m = self.mMali[identifier]
                outfile.write("%i\t%s\t%i\t%s\n" % (
                    m.mFrom+1, m.mString, m.mTo, m.getIdentifier()) )
                    
        elif format == "fasta":
            for identifier in self.mIdentifiers:
                m = self.mMali[identifier]
                if write_ranges:
                    outfile.write( ">%s/%i-%i\n%s\n" % (m.getIdentifier(), m.mFrom + 1, m.mTo, m.mString) )
                else:
                    outfile.write( ">%s\n%s\n" % (identifier, m.mString) )
                    
        elif format == "stockholm":
            outfile.write("# STOCKHOLM 1.0\n")

            if options:
                for o in options:
                    outfile.write( "%s\n" % o )
                    
            ## calculate offset:
            max_l = 0
            for identifier in self.mIdentifiers:
                m = self.mMali[identifier]
                id = m.getIdentifier()
                # tab does not work as separator
                if m.mTo and write_ranges:
                    x = "%s/%i-%i" % (id, m.mFrom+1, m.mTo)
                else:
                    x = "%s" % (id)
                max_l = max(max_l, len(x))
            for identifier in self.mAnnotations.keys():
                x = "#=GC %s" % identifier
                max_l = max(max_l, len(x))

            format = "%-" + str(max_l) + "s  %s\n"
            for identifier in self.mIdentifiers:
                m = self.mMali[identifier]
                id = m.getIdentifier()
                # tab does not work as separator
                if m.mTo and write_ranges:
                    x = "%s/%i-%i" % (id, m.mFrom+1, m.mTo)
                else:
                    x = "%s" % (id)

                outfile.write( format % (x, m.mString) )
                
            for identifier, value in self.mAnnotations.items():
                x = "#=GC %s" % identifier
                outfile.write( format % (x,value) )
            
            outfile.write("//\n")

        elif format == "phylip":

            outfile.write( "%i %i\n" % (self.getLength(), self.getWidth() ) )
            
            for identifier in self.mIdentifiers:
                outfile.write( "%s   %s\n" % (identifier, self.mMali[identifier].mString))

        elif format.lower() == "profile":
            if self.mName:
                name = self.mName
            else:
                name = ",".join(self.mIdentifiers)
                
            outfile.write( ">profile=%s length=%i width=%i\n" % (name, self.getWidth(), self.getLength() ) )

            for identifier in self.mIdentifiers:
                outfile.write( "%s\n" % (self.mMali[identifier].mString))

        elif format == "nexus":
            # nexus formatted output - MrBayes conformant.
            outfile.write( "#NEXUS\n" )
            outfile.write( "begin data;\n" )
            outfile.write( "  dimensions ntax=%i nchar=%i;\n" % (self.getLength(), self.getWidth()))
            outfile.write( "  format datatype=dna interleave=no gap=%s;\n" % (self.mGapChar))
            outfile.write( "  matrix\n" )
            max_len = max( map( lambda x: len(x), self.mIdentifiers) )
            format = "  %-" + str(max_len) + "s %s\n"
            for identifier in self.mIdentifiers:
                outfile.write( format % (identifier, self.mMali[identifier].mString) )
            outfile.write( "  ;\n" )
            outfile.write( "end;\n" )

        else:
            raise "unknown alignment format %s" % format


    def removeUnalignedEnds( self ):
        """remove unaligned ends in the multiple alignment.

        unaligned ends correspond to lower-case characters.
        """
        pattern_start = re.compile( "^([- .a-z]+)" )
        pattern_unaligned = re.compile("[a-z]")
        
        for s in self.mMali.values():

            t0 = time.time()
            first = pattern_start.match( s.mString )
            if first:
                first = first.groups()[0]
                nchars = len( pattern_unaligned.findall( first ) )
                s.mFrom += nchars
                s.mString = self.mGapChar * len(first) + s.mString[len(first):]

            t0 = time.time()
            ## search from the back end by reversing. This is much faster than
            ## using $ from the back.
            last = pattern_start.match( s.mString[::-1] )
            if last:
                last = last.groups()[0]
                nchars = len( pattern_unaligned.findall( last ) )
                s.mTo -= nchars
                l = len(s) - len(last)
                s.mString = s.mString[:l] + self.mGapChar * l

    def upperCase(self):
        """set all characters to upper case."""
        for k, s in self.mMali.items():
            s.mString = s.mString.upper()

    def lowerCase(self):
        """set all characters to lower case."""
        for k, s in self.mMali.items():
            s.mString = s.mString.lower()

    def removeEndGaps( self ):
        """remove end gaps.

        end gaps do not include any characters and thus
        the alignment coordinates won't change.
        """

        pattern_start_gaps = re.compile( "^([- ]+)" )
        
        min_from = self.mLength
        max_to = 0
        
        for s in self.mMali.values():

            first = pattern_start_gaps.match( s.mString )
            if first:
                first = first.groups()[0]
                min_from = min( min_from, len(first) )
                
            ## search from the back end by reversing. This is much faster than
            ## using $ from the back.
            last = pattern_start_gaps.search( s.mString[::-1] )
            if last:
                last = last.groups()[0]
                max_to = max( max_to, len(s) - len(last) )

        for s in self.mMali.values():
            s.mString = s.mString[min_from:max_to]

        self.mLength = min( map( lambda x: x.mString, self.mMali.values() ) )

    def insertColumns( self, position, num_gaps, keep_fixed = None, char = "-" ):
        """insert gaps at position into multiple alignment.

        if keep_constant is a list of identifiers, those are kept constant,
        instead, gaps are added to the end.
        """

        last_pos = min(self.getWidth(), position + num_gaps)
        for id, seq in self.mMali.items():
            if keep_fixed and id in keep_fixed:
                seq.insertColumns( last_pos, num_gaps, char )
            else:
                seq.insertColumns( position, num_gaps, char )

    def removeGaps( self, 
                    allowed_gaps = 0, 
                    minimum_gaps = 1, 
                    frame = 1 ):
        """remove gappy columns.

        allowed_gaps: number of gaps allowed for column to be kept
        minimum_gaps: number of gaps for column to be removed

        set minimum_gaps to the number of sequences to remove columns
        with all gaps.

        If frame is > 1 (3 most likely), then a whole codon is removed
        as soon as there is one column to be removed.
        """

        self.removePattern(
            match_function = lambda x: x in self.mGapChars,
            allowed_matches = allowed_gaps,
            minimum_matches = minimum_gaps,
            delete_frame = frame )

    def removePattern( self, 
                       match_function, 
                       allowed_matches = 0, 
                       minimum_matches = 1, 
                       delete_frame = 1,
                       search_frame = 1 ):
        """remove columns (or group of columns), that match a certain pattern.

        allowed_matches: number of matches allowed so that column is still kept
        minimum_matches: number of matches required for column to be removed

        set minimum_matches to the number of sequences to remove columns
        with all gaps.

        Patterns are matches in search_frame. For example, if frame is 3, 
        whole codons are supplied to match_function.

        delete_frame specifies the frame for deletion. If it is set to 3,
        codons are removed if already one column matches.

        Example: remove all columns that contain at least one stop-codon:

        removePattern( lambda x: x.upper() in ("TAG", "TAA", "TGA"),
                allowed_matches = 0,
                minimum_matches = 1,
                search_frame = 3,
                delete_frame = 3)
        """
        nmatches = [0] * self.getWidth()

        for s in map(lambda x: x.mString, self.mMali.values()):
            for x in range(0, len(s), search_frame):
                segment = s[x:x+search_frame]
                if match_function( segment ):
                    nmatches[x] += 1

        columns = []
        delete_columns = []
        
        for x in range( len(nmatches) ):
            if nmatches[x] >= allowed_matches and nmatches[x] < minimum_matches :
                columns.append(x)
            else:
                delete_columns.append(x)

        if delete_frame != 1:
            s = set(columns)
            for x in delete_columns:
                start = int( math.floor( float(x) / delete_frame ) * delete_frame )
                end = start + delete_frame
                for c in range(start,end):
                    if c in s:
                        s.remove(c)

            columns = list(s)
            columns.sort()
            
        self.takeColumns( columns )

    def removeEmptySequences( self ):
        """remove sequences that are completely empty.
        """
        new_ids = []
        for id in self.mIdentifiers:
            if self.countCharacters( self.mMali[id].mString ) == 0:
                del self.mMali[id]
                continue
            new_ids.append(id)
        self.mIdentifiers = new_ids

    def upper( self ):
        """convert all characters in mali to uppercase."""
        
        for s in self.mMali.values():
            s.mString = s.mString.upper()
            
    def lower( self ):
        """convert all characters in mali to lowercase."""
        
        for s in self.mMali.values():
            s.mString = s.mString.lower()

    def shiftAlignment( self, map_id2offset ):
        """shift alignment by offset."""

        for identifier, m in self.mMali.items():
            if identifier in map_id2offset:
                o = map_id2offset[identifier]
                m.mFrom += o
                m.mTo += o

    def markCodons( self, mode="case" ):
        """mark codons.
        """
        for identifier, m in self.mMali.items():
            s = m.mString
            if len(s) % 3 != 0:
                raise "sequence %s not divisible by 3" % (m.mId)

            is_upper = True
            sequence = []
            for x in range(0, len(s), 3):
                if is_upper:
                    sequence.append( s[x:x+3].upper() )
                    is_upper = False
                else:
                    sequence.append( s[x:x+3].lower() )
                    is_upper = True

            m.mString = "".join( sequence )
        
        
    def markTransitions( self, map_id2transitions, mode="case" ):
        """mark transitions in the multiple alignment.

        if mode == case, then upper/lower case is used for the transitions

        Otherwise, a character given by mode is inserted.
        """

        if mode in ("case", "keep-odd", "keep-even"):

            ## check, if the whole alignment needs to be masked/marked:
            if "mali" in map_id2transitions:
                transitions = map_id2transitions["mali"]
                for identifier, s in self.mMali.items():
                    new_chars = []
                    is_upper = True
                    is_first = False
                    
                    for c in range( len(s) ):

                        if c in transitions:
                            is_first = True
                            if is_upper:
                                is_upper = False
                            else:
                                is_upper = True

                        x = s.mString[c]
                        
                        if mode == "case":
                            if x in string.lowercase:
                                x = self.mMaskChar
                            if is_upper:
                                x = string.upper(x)
                            else:
                                x = string.lower(x)
                        elif mode == "keep-even":
                            if is_upper:
                                x = self.mGapChar
                        elif mode == "keep-odd":
                            if not is_upper:
                                x = self.mGapChar

                        new_chars.append( x )

                    s.mString = "".join(new_chars)

            ## now do individual sequences
            for identifier, s in self.mMali.items():
                if identifier not in map_id2transitions: continue

                new_chars = []
                c = s.mFrom
                
                is_upper = True
                is_first = False
                transitions = map_id2transitions[identifier]
                for x in s.mString:
                    if x in self.mGapChars:
                        pass
                    else:
                        if c in map_id2transitions[identifier]:
                            is_first = True
                            if is_upper:
                                is_upper = False
                            else:
                                is_upper = True
                        c += 1

                        if x in string.lowercase:
                            x = self.mMaskChar

                        if mode == "case":
                            if is_upper:
                                x = string.upper(x)
                            else:
                                x = string.lower(x)
                        elif mode == "keep-even":
                            if is_upper:
                                x = self.mGapChar
                        elif mode == "keep-odd":
                            if not is_upper:
                                x = self.mGapChar
                                
                    new_chars.append( x )

                s.mString = "".join(new_chars)
        else:
            raise "character insertion not implemented yet."

    def buildColumnMap( self, other, join_field = None ):
        """build map of columns in other to this."""

        if not join_field:
            join_field = other.mIdentifiers[0]
            
        if join_field not in other.mMali or \
           join_field not in self.mMali:
            raise "line %s not in both alignments." % (join_field)

        this_seq = self.mMali[join_field]
        other_seq = other.mMali[join_field]

        if this_seq.mFrom != other_seq.mFrom or \
           this_seq.mTo != other_seq.mTo:
            raise "residue ranges for sequence %s doe not correspond." % (join_field)

        map_this2other = []

        this_seq = this_seq.mString.upper()
        other_seq = other_seq.mString.upper()        

        # position in other
        o = 0
        for c in this_seq:
            if c in self.mGapChars:
                map_this2other.append( None )
            else:
                while other_seq[o] != c:
                    o += 1
                map_this2other.append( o )
                o += 1
        return map_this2other

    def shuffle( self, frame = 1 ):
        """shuffle multiple alignment.

        The frame determines the block size for shuffling. Use 3 for codons in
        a multiple alignment without frame-shifts.
        """
        columns = range( self.getNumColumns() // frame )
        random.shuffle( columns )
        if frame > 1:
            cc = []
            for x in columns: cc += [ x * frame + y for y in range( frame ) ]
            columns = cc
        self.takeColumns( columns )

    def propagateMasks( self, min_chars = 1, mask_char = "x" ):
        """propagate masked characters to all rows of a multiple alignment
        within a column.

        If there is at least min_chars in a mali column, that are masks,
        propagate the masks to all other rows.
        """

        masks_per_column = {}
        for identifier, s in self.mMali.items():
            r = s.mString.lower()
            for x in range(len(r)):
                if r[x] == mask_char:
                    if x not in masks_per_column: masks_per_column[x] = 0
                    masks_per_column[x] += 1

        columns_to_mask = []
        for c, n in masks_per_column.items():
            if n >= min_chars:
                columns_to_mask.append( c )
                
        columns_to_mask.sort()
        
        self.maskColumns( columns_to_mask, mask_char = mask_char )

    def propagateTransitions( self, min_chars = 1 ):
        """propagate lower case in a column to all residues.
        """
        columns_to_change = set()
        for identifier, s in self.mMali.items():
            r = s.mString
            for x in range(len(r)):
                if r[x] in string.lowercase:
                    columns_to_change.add( x )

        columns_to_change = list(columns_to_change)
        columns_to_change.sort()
        
        self.mapColumns( columns_to_change, string.lower )

    def takeColumns( self, columns ):
        """restrict alignments to certain columns."""
        for identifier, s in self.mMali.items():
            s.takeColumns( columns )
            
        for key, anno in self.mAnnotations.items():
            self.mAnnotations[key] = "".join( [ anno[c] for c in columns ] )

    def maskColumns( self, columns, mask_char = "x"):
        """mask columns in a multiple alignment."""

        for identifier, s in self.mMali.items():
            s.maskColumns( columns, mask_char = mask_char )

    def mapColumns( self, columns, map_function):
        """apply map_function to all residues in columns."""

        for identifier, s in self.mMali.items():
            s.mapColumns( columns, map_function )

    def recount( self, reset_first = False ):
        """recount residue in alignments."""
        for id, seq in self.mMali.items():
            if reset_first: seq.mFrom = 0
            seq.mTo = seq.mFrom + self.countCharacters( seq.mString )

    def maskColumn( self, column, mask_char = "x"):
        """mask a column."""
        
        for identifier, s in self.mMali.items():
            s.maskColumn(column, mask_char)
                
    def copyAnnotations( self, other ):
        """copy annotations from annother mali."""

        map_this2other = self.buildColumnMap( other )
        ncols = self.getWidth()
        
        for key, annotation in other.mAnnotations.items():
            a = []
            for x in range(ncols):
                m = map_this2other[x] 
                if m !=  None:
                    a.append( annotation[m] )
                else:
                    a.append( self.mGapChar )
            
            self.mAnnotations[key] = "".join(a)
            
    def getAnnotation( self, key ):
        """return annotation associated with key."""
        return self.mAnnotations[key]

    def setAnnotation( self, key, value ):
        """set annotation associated with key to value."""
        self.mAnnotations[key] = value

    def addAnnotation( self, key, annotation ):
        """add annotation."""
        xkey = key
        n = 0
        while xkey in self.mAnnotations:
            xkey = "%s_%i" % (key, n)
            
        self.mAnnotations[xkey] = re.sub( "\s", "", annotation )

    def truncate( self, first, last):
        """truncate alignment within range."""
        for key, value in self.mMali.items():
            value.truncate( first, last )

    def mapIdentifiers( self, map_old2new = None, pattern_identifier="ID%06i" ):
        """map identifiers in multiple aligment.

        if map_old2new is not given, a new map is created (map_new2old)
        """
        new_identifiers = []
        if map_old2new != None:
            for id in self.mIdentifiers:
                new_id = map_old2new[id]
                new_identifiers.append( new_id )
                
                entry = self.mMali[id]
                del self.mMali[id]                
                # entry.mId = new_id
                self.mMali[new_id] = entry
        else:
            map_old2new = {}
            n = 1
            for id in self.mIdentifiers:
                new_id = pattern_identifier % n
                new_identifiers.append( new_id )                
                
                entry = self.mMali[id]
                del self.mMali[id]                
                # entry.mId = new_id
                self.mMali[new_id] = entry

                map_old2new[new_id] = id
                n += 1
                
        self.mIdentifiers = new_identifiers

        return map_old2new

    def getAlphabet( self ):
        """get alphabet from the multiple alignment.

        Alphabet is "na", if more than 90% of characaters are "actgxn",
        otherwise it is "aa".
        """
        s = "".join( map( lambda x: x.mString, self.values())).lower()
        s = re.sub( "[%s]" % "".join(self.mGapChars), "", s )
        ss = re.sub( "[acgtxn]", "", s)
        if float(len(ss)) < (len(s) * 0.1):
            return "na"
        else:
            return "aa"

    def checkLength( self ):
        """check lengths of aligned strings.

        Return false if they are inconsistent."""

        l = None
        for s in self.mMali.values():
            if l == None:
                l = len(s.mString)
            else:
                if l != len(s.mString):
                    return False
        return True

    def clipByAnnotation( self, key, chars = ""):
        """restrict alignment to positions where 
        annotation identified by key in chars.

        if chars is empty, nothing is clipped.
        """
        if key not in self.mAnnotations: return
        if chars == "": return
        if self.getNumColumns() == 0: return

        anno = self.mAnnotations[key]
        columns = []
        
        for x in range( len( anno )):
            if anno[x] in chars:
                columns.append( x )
        
        self.takeColumns( columns )

    def apply( self, f ):
        """apply function f to every row in the multiple alignment."""
        for s in self.mMali.values():
            f(s)

    def filter( self, f ):
        '''filter multiple alignment using function *f*.

        The function *f* should return True for entries that 
        should be kept and False for those that should be removed.
        '''
        ids, vals = [], []
        for id in self.mIdentifiers:
            s = self.mMali[id]
            if f(s):
                vals.append(s)
                ids.append( id )
        self.mIdentifiers = ids
        self.mMali = vals

class SequenceCollection( Mali ):
    """reads in a sequence collection, but permits
    several entries per id. 

    Note that this might cause problems with interleaved 
    formats like phylips or clustal.

    This mapping is achieved by appending a numeric
    suffix. The suffix is retained for the life-time
    of the object, but not output to a file.
    """

    
    def __init__(self):
        Mali.__init__(self)

    def addEntry( self, s ):
        """add an aligned string object."""
        id = s.mId
        if id in self.mMali.keys():
            x = 1
            while "%s_%i" % (id, x) in self.mMali.keys():
                x += 1
                
            id = "%s_%i" % (id, x)
        self.mIdentifiers.append( id )
        s.mIsMangled = True

        self.mMali[id] = s

    def readFromFile( self, infile, format = "fasta" ):
        """read multiple alignment from file in various format."""
        
        self.mMali = {}
        self.mIdentifiers = []

        pattern_parse_ranges=re.compile("(\S+)/(\d+)-(\d+)")

        if type(infile) == types.ListType or type(infile) == types.TupleType:
            lines = infile
        else:
            lines = infile.readlines()

        if format not in ("stockholm"):
            # save comments 
            self.mComments = filter( lambda x: x[0] == "#", lines)
            lines = filter( lambda x: x[0] != "#", lines)
        else:
            self.mComments = []

        def getId( id, s ):
            x = pattern_parse_ranges.match( id )
            if x:
                id, fr, to = x.groups()
                fr, to = int(fr) - 1, int(to)
            else:
                fr, to = 0, self.countCharacters( s )
                self.mWriteRanges = False

            return id, fr, to

        ###############################################################################
        if format.lower() == "plain":

            for line in lines:
                if not line.strip(): continue
                data = line[:-1].split("\t")
                id = data[3]
                xid = id
                x = 0
                while xid in self.mMali:
                    xid = id + "-" + str(x)
                    x += 1

                self.addEntry( AlignedString(xid, int(data[0]) - 1, int(data[2]), data[1] ) )

        ###############################################################################
        elif format.lower() == "fasta":
            pattern_identifier = "\S+"
            id = None
            fragments = []
            for line in lines:
                if line[0] == ">":
                    if id:
                        s = re.sub( "\s", "", string.join( fragments, ""))
                        id, fr, to = getId( id, s )
                        self.addEntry( AlignedString(id, fr, to, s) )
                        
                    id = re.search( "^(%s)" % pattern_identifier, line[1:-1]).group(0)
                    fragments = []
                    continue
                fragments.append( line[:-1] )

            s = re.sub( "\s", "", string.join( fragments, ""))                
            id, fr, to = getId( id, s )
            self.addEntry( AlignedString(id, fr, to, s) )
            
        ###############################################################################
        elif format.lower() == "phylip":
            raise "phylip not implemented"

        ###############################################################################            
        elif format.lower() == "clustal":
            raise "clustal not implemented"

        elif format.lower() == "stockholm":
            raise "stockholm not implemented"
        
        if len(self.mMali) == 0:
            self.mLength = 0
        else:
            self.mLength = min( map( lambda x: len(x.mString), self.mMali.values() ) )

    
###############################################################################            
def convertMali2Alignlib( mali ):
    '''convert a multiple alignment of type :class:`Mali`
    into an alignlib multiple alignment object.
    '''

    import alignlib
    m = alignlib.makeMultipleAlignment()
    for identifier in mali.getIdentifiers():
        a = alignlib.makeAlignatum( mali[identifier] )
        m.add( a )
    return m

###############################################################################            
def convertAlignlib2Mali( mali, identifiers = None, seqs = None ):
    """convert a multiple alignment into an alignlib multiple alignment object."""
    m = Mali()

    if not identifiers:
        identifiers = [ "%i" % x for x in range( mali.getNumSequences() ) ]

    if seqs == None:
        # old style MultipleAlignment
        for x in range(mali.getNumSequences()):
            a = mali.getRow( x )
            m.addSequence( identifiers[x], a.getFrom(), a.getTo(), a.getString() )
    else:
        import alignlib
        output = alignlib.MultAlignmentFormatPlain( mali, seqs )
        for x in range(mali.getNumSequences()):
            a = output.mData[ x ]
            m.addSequence( identifiers[x], a.getFrom(), a.getTo(), a.getString() )
        
    return m




#cimport csamtools

from pysam.csamtools cimport *

import collections, array, struct
import CGAT.Experiment as E

FLAGS = {
    1: 'paired',
    2: 'proper_pair',
    4: 'unmapped',
    8: 'mate_unmapped',
    16: 'reverse',
    32: 'mate_reverse',
    64: 'read1',
    128: 'read2',
    256: 'secondary',
    512: 'qc_fail',
    1024: 'duplicate'  }

def count( Samfile samfile,
           remove_rna,
           rna):
    '''

    '''
    cdef bint _remove_rna = remove_rna

    cdef AlignedRead read

    # counters
    cdef int ninput = 0
    cdef int nduplicates = 0
    cdef int nrna = 0
    cdef int nfiltered = 0

    cdef int max_hi = 0
    # count nh, nm tags
    nh_filtered, nm_filtered = collections.defaultdict( int ), collections.defaultdict( int )
    nh_all, nm_all = collections.defaultdict( int ), collections.defaultdict( int )
    mapq_filtered, mapq_all = collections.defaultdict( int ), collections.defaultdict( int )

    cdef int * flags_counts = <int*>calloc( len(FLAGS), sizeof(int) )

    # helper variables
    cdef int last_tid = -1
    cdef int last_pos = 0

    duplicates = collections.defaultdict( int )
    counts = collections.defaultdict( int )
    cdef int count = 0

    cdef int tid = -1
    cdef int flag

    cdef uint8_t * v
    cdef int32_t nm
    cdef int32_t nh
    cdef int32_t hi
    cdef int x
    cdef int lflags = len(FLAGS)
    cdef int f

    for read in samfile:

        flag = read._delegate.core.flag

        # is paired and read2
        #if flag & 1 and flag & 128:
            # ignore second reads to avoid double counting pairs
            # this needs to be done properly by counting in
            # pairs
        #    continue

        ninput += 1

        f = 1
        for x from 0 <= x < lflags:
            if flag & f: flags_counts[x] += 1
            f = f << 1

        # get maximum NI field
        v = bam_aux_get(read._delegate, 'HI')
        if v != NULL:
            hi = <int32_t>bam_aux2i(v)
            if hi > max_hi: max_hi = hi

        v = bam_aux_get(read._delegate, 'NH')
        if v != NULL:
            nh = <int32_t>bam_aux2i(v)
            nh_all[nh] += 1
        else:
            nh = -1

        v = bam_aux_get(read._delegate, 'NM')
        if v != NULL:
            nm = <int32_t>bam_aux2i(v)
            nm_all[nm] += 1
        else:
            nm = -1

        mapq_all[read.mapq] += 1

        # skip unmapped reads
        if read._delegate.core.flag & 4: continue

        if read.tid != last_tid:
            contig = samfile.getrname( read.rname )

        # note: does not take into account gaps within reads
        # or partial overlap.
        if rna and rna.contains( contig, read.pos, read.pos + read.qlen ):
            nrna += 1
            if _remove_rna: continue
        
        nfiltered += 1

        if nh >= 0: nh_filtered[nh] += 1
        if nm >= 0: nm_filtered[nm] += 1
        mapq_filtered[read.mapq] += 1

        # duplicate analysis - simply count per start position
        # ignoring sequence and strand
        if read.tid == last_tid and read.pos == last_pos:
            count += 1
            nduplicates += 1
            continue

        if count > 1:
            counts[count] += 1

        count = 1
        last_tid, last_pos = read.tid, read.pos

    c = E.Counter()
    
    c.input = ninput
    c.filtered = nfiltered
    c.rna = nrna
    c.duplicates = nduplicates

    # convert flags to labels
    t = {}
    f = 1
    for x from 0 <= x < lflags:
        t[FLAGS[f]] = flags_counts[x]
        f = f << 1

    return c, t, nh_filtered, nh_all, nm_filtered, nm_all, mapq_filtered, mapq_all, max_hi
